/*
*  Copyright (c) 2023 NetApp
*  All rights reserved
 */

package utils

import (
	appsv1 "k8s.io/api/apps/v1"
)

// Labels / annotations used to dictate proxy functionality.
// See https://istio.io/latest/docs/reference/config/annotations/ for more details
const (
	// The revision label the webhook selectors select for
	RevisionLabel = "istio.io/rev"

	// ActiveIstiodTagIdentifier is the identifier associated with the
	// control plane revision that is currently active.
	ActiveIstiodTagIdentifier = "default"

	ProxyMemoryRequestsLabel = "sidecar.istio.io/proxyMemory"
	ProxyMemoryLimitsLabel   = "sidecar.istio.io/proxyMemoryLimit"
	ProxyCPURequestsLabel    = "sidecar.istio.io/proxyCPU"
	ProxyCPULimitsLabel      = "sidecar.istio.io/proxyCPULimit"

	// We use 'include' logic here instead of 'exclude' as we only want CSM and DSMF
	// communication to be proxied.
	// A comma separated list of inbound ports to be included from redirection to Envoy
	IncludeInboundPortsLabel = "traffic.sidecar.istio.io/includeInboundPorts"
	// A comma separated list of outbound ports to be included from redirection to Envoy
	IncludeOutboundPortsLabel = "traffic.sidecar.istio.io/includeOutboundPorts"

	CSMAndDSMFPorts = "11104,11105,7700"

	// We chain 'includeOutboundPortsLabel' with this to get the correct 'inclusive-only'
	// functionality we want.
	// See last comment on https://github.com/istio/istio/issues/32677 for more details
	IncludeOutboundIPRangesLabel = "traffic.sidecar.istio.io/includeOutboundIPRanges"

	Localhost = "127.0.0.1/32"

	// Overrides for the proxy configuration for this specific proxy
	ProxyConfigLabel = "proxy.istio.io/config"

	// proxyMetadata.ISTIO_META_DNS_CAPTURE makes DNS traffic filter through the istio-proxy first if 'true'.
	// If istio-proxy cannot resolve the query it will forward the traffic to whatever is in its /etc/resolv.conf.
	// This config breaks secd DNS and thus secd as a feature. We want to explicitly set this to false to
	// maintain any change in ASM default functionality doesn't break secd.
	ProxyConfig = `{ "proxyMetadata": { "ISTIO_META_DNS_CAPTURE": "false" }}`
)

// AddLabelsAnnotations
func AddLabelsAnnotations(deployment *appsv1.Deployment) {
	if deployment.Spec.Template.Labels == nil {
		deployment.Spec.Template.Labels = make(map[string]string)
	}
	if deployment.Spec.Template.Annotations == nil {
		deployment.Spec.Template.Annotations = make(map[string]string)
	}

	deployment.Spec.Template.Labels[RevisionLabel] = ActiveIstiodTagIdentifier
	deployment.Spec.Template.Annotations[IncludeInboundPortsLabel] = CSMAndDSMFPorts
	deployment.Spec.Template.Annotations[IncludeOutboundPortsLabel] = CSMAndDSMFPorts
	deployment.Spec.Template.Annotations[IncludeOutboundIPRangesLabel] = Localhost
	deployment.Spec.Template.Annotations[ProxyConfigLabel] = ProxyConfig
}

// unsetIstioDataplaneFields
func DeleteLabelsAnnotations(deployment *appsv1.Deployment) {
	delete(deployment.Spec.Template.Labels, RevisionLabel)
	delete(deployment.Spec.Template.Annotations, IncludeInboundPortsLabel)
	delete(deployment.Spec.Template.Annotations, IncludeOutboundPortsLabel)
	delete(deployment.Spec.Template.Annotations, IncludeOutboundIPRangesLabel)
	delete(deployment.Spec.Template.Annotations, ProxyConfigLabel)
}
