package v2

import (
	"context"
	"fmt"

	"quark.netapp.io/volume-controller/pkg/quarkconfig"

	apierrors "k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	log "github.com/sirupsen/logrus"
)

// ServiceMeshAnthosCleanup deletes ASM control plane resources. This API should only
// be called after the fleet mesh for the cluster has been updated to 'manual' mode. Otherwise,
// these resources will be re-created by google managed controllers.
//
// The steps here should always be inline with:
// https://cloud.google.com/service-mesh/docs/uninstall#uninstall
//
// TODO: There have been a couple cases so far where these steps have change on us. We may
// want to consider a more dynamic approach like maintaining a configmap of resources
// to delete or a helm chart. Right now, when the steps change we are force to generate
// a new SDE build for the new NVCClient code. This is contributing to our already
// heavy processes.
func (client *NVCClient) ServiceMeshAnthosCleanup(ctx context.Context) error {
	var err error
	validatingWebhookName := "istiod-istio-system-mcp"
	err = client.GetKubeClient().AdmissionregistrationV1().
		ValidatingWebhookConfigurations().
		Delete(ctx, validatingWebhookName, metav1.DeleteOptions{})
	if err != nil && !apierrors.IsNotFound(err) {
		return fmt.Errorf("error deleting validating webhook '%v': %v", validatingWebhookName, err)
	}
	log.Debugf("Successfully deleted validatingwebhookconfiguration '%v'", validatingWebhookName)

	err = client.GetKubeClient().AdmissionregistrationV1().
		MutatingWebhookConfigurations().
		DeleteCollection(ctx, metav1.DeleteOptions{}, metav1.ListOptions{
			LabelSelector: quarkconfig.RevisionLabel,
		})
	if err != nil {
		return fmt.Errorf("error deleting all service mesh mutating webhooks with selector '%v': %v",
			quarkconfig.RevisionLabel, err)
	}
	log.Debugf("Successfully deleted all mutatingwebhookconfigurations with label '%v'", quarkconfig.RevisionLabel)

	deleteNamespace := func(namespace string) error {
		err := client.GetKubeClient().CoreV1().
			Namespaces().
			Delete(ctx, namespace, metav1.DeleteOptions{})
		if err != nil && !apierrors.IsNotFound(err) {
			return fmt.Errorf("error deleting '%v' namespace: %v", namespace, err)
		}
		log.Debugf("Successfully deleted namespace '%v'", namespace)
		return nil
	}
	if err := deleteNamespace(quarkconfig.IstioSystemNamespace); err != nil {
		return err
	}
	if err := deleteNamespace(quarkconfig.ASMSystemNamespace); err != nil {
		return err
	}

	resource := "mdp-controller"
	err = client.GetKubeClient().AppsV1().
		Deployments(quarkconfig.KubeSystemNamespace).
		Delete(ctx, resource, metav1.DeleteOptions{})
	if err != nil && !apierrors.IsNotFound(err) {
		return fmt.Errorf("error deleting deployment '%v/%v': %v", quarkconfig.KubeSystemNamespace, resource, err)
	}
	log.Debugf("Successfully deleted deployment '%v/%v'", quarkconfig.KubeSystemNamespace, resource)

	// The anthos uninstall steps explicitly say to delete the "istio-cni-plugin-config". In older versions there
	// is also an "istio-cni-config" which is likely the predecsor to the "istio-cni-plugin-config"-map as of the following PR:
	// https://github.com/GoogleCloudPlatform/netd/pull/251
	// The istio-cni-config was never included in the cleanup steps. Despite that, we should probably delete this
	// CM too as it is an artifact of the ASM install.
	for _, resource := range []string{"istio-cni-plugin-config", "istio-cni-config"} {
		err = client.GetKubeClient().CoreV1().
			ConfigMaps(quarkconfig.KubeSystemNamespace).
			Delete(ctx, resource, metav1.DeleteOptions{})
		if err != nil && !apierrors.IsNotFound(err) {
			return fmt.Errorf("error deleting configmap '%v/%v': %v", quarkconfig.KubeSystemNamespace, resource, err)
		}
	}

	resource = "istio-cni-node"
	err = client.GetKubeClient().AppsV1().
		DaemonSets(quarkconfig.KubeSystemNamespace).
		Delete(ctx, resource, metav1.DeleteOptions{})
	if err != nil && !apierrors.IsNotFound(err) {
		return fmt.Errorf("error deleting daemonset '%v/%v': %v", quarkconfig.KubeSystemNamespace, resource, err)
	}
	log.Debugf("Successfully deleted daemonset '%v/%v'", quarkconfig.KubeSystemNamespace, resource)
	return nil
}
