package v2

import (
	"context"
	"fmt"
	"strings"

	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/runtime/schema"

	_ "k8s.io/client-go/plugin/pkg/client/auth"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"

	"quark.netapp.io/volume-controller/pkg/nvc"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

var (
	// supporting smaller storagePool for upgrade and internal configs
	minStoragePoolSize      resource.Quantity = resource.MustParse("1Gi")
	maxStoragePoolSize      resource.Quantity = resource.MustParse("200Ti")
	maxStoragePoolSizeAzure resource.Quantity = resource.MustParse("100Ti")
)

const (
	versionSPUpgradeErr = "Failed to upgrade NetAppStoragePool"
	upgradeSPAbortErr   = "Failed to abort NetAppStoragePoolUpgrade"
	CMEKConfigMapErr    = "Failed to create cmek config map: %s not provided"
	encryptErr          = "Failed to encrypt NetAppStoragePool"
	encryptPauseErr     = "Failed to pause NetAppStoragePool encryption"
	encryptResumeErr    = "Failed to resume NetAppStoragePool encryption"
)

func isValidCMEKLogLevel(level string) bool {
	logLevel := strings.ToLower(level)
	CMEKLogLevels := []string{"debug", "info", "notice", "warning", "error", "critical", "alert", "emergency"}
	for _, value := range CMEKLogLevels {
		if value == logLevel {
			return true
		}
	}
	return false
}

type MigrateZoneAbortZoneFailureError struct {
}

func (e MigrateZoneAbortZoneFailureError) Error() string {
	return "Unable to abort migration caused by zone failure."
}

type MigrateZoneAbortNoMigrationFailureError struct {
}

func (e MigrateZoneAbortNoMigrationFailureError) Error() string {
	return "No zone migration to abort."
}

type GlobalAccessUpdateAbortNoUpdateError struct{}

func (e GlobalAccessUpdateAbortNoUpdateError) Error() string {
	return "global access update not in progress"
}

type GlobalAccessAbortInProgressError struct{}

func (e GlobalAccessAbortInProgressError) Error() string {
	return "global access abort in progress"
}

// NetAppStoragePoolOnline returns "true" when the Quark pod is ready to serve data.
func (client *NVCClient) NetAppStoragePoolOnline(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsNetAppStoragePoolOnline(storagePool)
}

// NetAppStoragePoolDisabled returns "true" when the SP is disabled.
func (client *NVCClient) NetAppStoragePoolDisabled(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return storagePool.Status.Disabled
}

// NetAppStoragePoolSMBConfigured returns "true" if SMB server id configured in storagepool.
func (client *NVCClient) NetAppStoragePoolSMBConfigured(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsStoragePoolSMBConfigured(storagePool)
}

// NetAppStoragePoolSMBReady Returns "true" if SMB is configured and ready for use in storagepool.
func (client *NVCClient) NetAppStoragePoolSMBReady(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsStoragePoolSMBConfigured(storagePool) && util.IsSMBServerConfigurationEqual(storagePool)
}

// NetAppStoragePoolSMBInitialized Returns "true" if SMB is configured and ready for use in storagepool.
func (client *NVCClient) NetAppStoragePoolSMBInitialized(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsStoragePoolSMBInitialized(storagePool)
}

// NetAppStoragePoolLdapReady Returns "true" if LDAP is configured and ready for use in storagepool.
func (client *NVCClient) NetAppStoragePoolLdapReady(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	if !client.EnableCCPd || !client.EnableLdap {
		return false
	}

	return (client.NetAppStoragePoolSMBReady(storagePool) && util.IsStoragePoolLdapReady(storagePool))
}

// Get the storage pool status
func (client *NVCClient) NetAppStoragePoolStatus(storagePoolUUID string) (bool, string, *quarkv1alpha1.NetAppStoragePool, error) {
	storagePoolLifeCycleState := false
	storagePoolLifeCycleStateDetails := "Storagepool is in error state"
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		storagePoolLifeCycleStateDetails = "Storagepool doesn't exists"
	} else if err != nil {
		storagePoolLifeCycleStateDetails = "Storagepool get failed"
	} else if client.NetAppStoragePoolOnline(storagePool) {
		storagePoolLifeCycleState = true
		storagePoolLifeCycleStateDetails = "Storagepool is available"
	}
	return storagePoolLifeCycleState, storagePoolLifeCycleStateDetails, storagePool, err
}

// Initiate NetAppStoragePool version upgrade
func (client *NVCClient) NetAppStoragePoolUpgrade(storagePoolUUID, version string) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetappStoragePool %s was not found for upgrade", storagePoolUUID)
	}
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("%s: %s", versionSPUpgradeErr, err)
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("%s: no quarks found", versionSPUpgradeErr)
	}
	quark := quarks[0]
	if quark.Status.Version != version {
		return nil, fmt.Errorf("%s: quark is not updated to version %s", versionSPUpgradeErr, version)
	}

	// initiate upgrade when Quark is Ready and updated with required version for storagePool upgrade
	log.WithFields(log.Fields{
		"currentVersion": storagePool.Status.QuarkVersion,
		"upgradeVersion": version,
	}).Debug("Initiating storagepool upgrade.")
	storagePool.Spec.QuarkVersion = version
	newStoragePool, err := client.netAppStoragePoolUpdate(storagePool)
	return newStoragePool, err
}

// Abort NetAppStoragePool version upgrade
func (client *NVCClient) NetAppStoragePoolUpgradeAbort(storagePoolUUID string) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetAppStoragePool %s was not found for aborting upgrade", storagePoolUUID)
	}
	if storagePool.Spec.QuarkVersion == "" {
		return nil, fmt.Errorf("%s: NetAppStoragePool Spec version is missing", upgradeSPAbortErr)
	}
	// check if abort is possible
	if storagePool.Status.QuarkVersion != "" && storagePool.Spec.QuarkVersion != storagePool.Status.QuarkVersion {
		err := client.checkUpgradeAbortStoragePool(storagePool.Status.Conditions)
		if err != nil {
			return nil, fmt.Errorf("%s: %s", upgradeSPAbortErr, err.Error())
		}
	}
	// revert to old version, in order to signal abort
	storagePool.Spec.QuarkVersion = storagePool.Status.QuarkVersion
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		_, staleError := err.(StaleVersionError)
		if staleError {
			return nil, err
		}
		return nil, fmt.Errorf("%s: %s", upgradeSPAbortErr, err.Error())
	}
	return storagePool, nil
}

// returns error if abort cannot proceed
func (client *NVCClient) checkUpgradeAbortStoragePool(conditions []quarkv1alpha1.NetAppStoragePoolCondition) error {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltSucceeded {
			// do not proceed with abort if halt succeeded
			return fmt.Errorf("deployment is being updated for storagePool upgrade")
		}
	}
	return nil
}

func (client *NVCClient) NetAppStoragePoolGet(storagePoolUUID string) (*quarkv1alpha1.NetAppStoragePool, error) {
	pool := &quarkv1alpha1.NetAppStoragePool{
		ObjectMeta: metav1.ObjectMeta{
			Name: storagePoolUUID,
		},
	}
	result, err := client.getWithRetry(pool)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppStoragePool), nil
}

func (client *NVCClient) NetAppStoragePoolGetVersion(storagePoolUUID string) (string, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return "", fmt.Errorf("failed to get the NetAppStoragePool version: %v", err)
	}
	return storagePool.Status.QuarkVersion, nil
}

// returns conditions from NetAppStoragePool CR status
func (client *NVCClient) NetAppStoragePoolGetConditions(storagePoolUUID string) ([]quarkv1alpha1.NetAppStoragePoolCondition, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
	}
	return storagePool.Status.Conditions, nil
}

func (client *NVCClient) NetAppStoragePoolGetAll() ([]*quarkv1alpha1.NetAppStoragePool, error) {
	storagePools := make([]*quarkv1alpha1.NetAppStoragePool, 0)
	result, err := client.listWithRetry(&quarkv1alpha1.NetAppStoragePoolList{}, metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	storagePoolsList := result.(*quarkv1alpha1.NetAppStoragePoolList)
	for i := range storagePoolsList.Items {
		storagePools = append(storagePools, &storagePoolsList.Items[i])
	}
	return storagePools, err
}

func (client *NVCClient) NetAppStoragePoolGetAvailableSpace(storagePoolUUID string) (availableSpace int64, err error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil || storagePool == nil {
		return availableSpace, fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
	}
	if !client.NetAppStoragePoolOnline(storagePool) {
		return availableSpace, fmt.Errorf("NetAppStoragePool %s is not online yet", storagePoolUUID)
	}
	volumes, err := client.NetAppVolumeListGetByStoragePool(storagePoolUUID)
	if err != nil {
		return availableSpace, err
	}
	var usedStoragePoolSize int64 = 0
	for _, volume := range volumes.Items {
		usedStoragePoolSize = usedStoragePoolSize + volume.Status.Size.Value()
	}
	availableSpace = storagePool.Status.Size.Value() - usedStoragePoolSize
	return availableSpace, nil
}

func (client *NVCClient) NetAppStoragePoolGetSMBStatus(storagePoolUUID string) (configured bool, err error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil || storagePool == nil {
		return false, fmt.Errorf("failed to get the NetAppStoragePool: %v", err)
	}
	if !client.NetAppStoragePoolOnline(storagePool) {
		return false, fmt.Errorf("NetAppStoragePool %s is not online yet", storagePoolUUID)
	}
	return client.NetAppStoragePoolSMBConfigured(storagePool), nil
}

func (client *NVCClient) NetAppStoragePoolGetAllSMB() ([]*quarkv1alpha1.NetAppStoragePool, error) {
	storagePools := make([]*quarkv1alpha1.NetAppStoragePool, 0)
	result, err := client.listWithRetry(&quarkv1alpha1.NetAppStoragePoolList{}, metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	storagePoolsList := result.(*quarkv1alpha1.NetAppStoragePoolList)
	for i := range storagePoolsList.Items {
		if storagePoolsList.Items[i].Spec.SMBServerConfiguration.Netbios != "" {
			storagePools = append(storagePools, &storagePoolsList.Items[i])
		}
	}
	return storagePools, err
}

func (client *NVCClient) NetAppVolumeListGetByStoragePool(storagePoolUUID string) (*quarkv1alpha1.NetAppVolumeList, error) {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.StoragePoolLabel] = storagePoolUUID
	labelSelector := fmt.Sprintf("quark.netapp.io/storagePool=%s", storagePoolUUID)
	listOptions := metav1.ListOptions{
		LabelSelector: labelSelector,
	}
	result, err := client.listWithRetry(&quarkv1alpha1.NetAppVolumeList{}, listOptions)
	if err != nil {
		return nil, err
	}
	volumesList := result.(*quarkv1alpha1.NetAppVolumeList)
	return volumesList, err
}

func (client *NVCClient) NetAppStoragePoolDelete(storagePoolUUID string, user, password string) error {
	volumes, err := client.NetAppVolumeListGetByStoragePool(storagePoolUUID)
	if err != nil {
		return err
	}
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return fmt.Errorf("failed to get the NetAppStoragePool '%v': %v", storagePoolUUID, err)
	}

	// prevent active/in-use SPs from getting deleted
	if len(volumes.Items) > 0 && !storagePool.Status.Disabled {
		return fmt.Errorf("delete request denied for NetAppStoragePool '%v', still in use", storagePool.Name)
	}

	// Check if this storage pool is in any peering relationships and prevent deletion if it is.
	// (EnableCCPd should be true, else peering CRD won't be available on the cluster)
	if client.EnableCRR {
		if clusterPeeringsList, err := client.NetAppClusterPeeringFilterByStoragePool(storagePool.Name); err != nil {
			return fmt.Errorf("failed to list NetAppClusterPeerings to check if storage pool %v is in a peering relationship: %v", storagePool.Name, err.Error())
		} else {
			// If storage pool is disabled, peering CRs will exist for it but we won't block deletion
			if len(clusterPeeringsList) > 0 {
				if !storagePool.Status.Disabled {
					return fmt.Errorf("delete request for NetAppStoragePool '%v' denied. %v NetAppClusterPeering resources are associated with this storage pool. "+
						"Delete all NetAppClusterPeering CRs associated with the NetAppStoragePool before deleting it.", storagePool.Name, len(clusterPeeringsList))
				} else {
					log.WithFields(log.Fields{
						"storagePool": storagePool.Name,
						"disabled":    storagePool.Spec.Disabled,
					}).Debug("storage pool in disabled state. Ignoring associated peering CR and proceeding with delete request")
				}
			}
		}
	}

	if storagePool.Spec.SMBServerConfiguration.Netbios != "" {
		storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
		if err != nil {
			return fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
		}
		_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, util.GetVolDeleteADSecretName(storagePool.Name), user, password)
		if err != nil {
			return err
		}
	}

	if storagePool.Spec.KerbServerConfiguration != nil &&
		storagePool.Spec.KerbServerConfiguration.KDCServerIP != "" &&
		storagePool.Spec.KerbServerConfiguration.ADServerName != "" {
		storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
		if err != nil {
			return fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
		}
		secretName := util.GetVolDeleteADSecretName(storagePool.Name + "-kerb-")
		_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
		if err != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"secret":      secretName,
			}).Error("Failed to create AD secret for kerberos delete.")
			return fmt.Errorf("failed to create AD secret for kerberos delete: %s", err.Error())
		}
		log.Debug("successfully created kerb delete secret: ", secretName)
	}

	return client.deleteWithRetry(storagePool)
}

func (client *NVCClient) netAppStoragePoolUpdate(storagePool *quarkv1alpha1.NetAppStoragePool) (*quarkv1alpha1.NetAppStoragePool, error) {
	result, err := client.updateWithRetry(storagePool)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppStoragePool), nil
}

// NetAppStoragePoolResizeAbort aborts the current StoragePool-Resize operation
func (client *NVCClient) NetAppStoragePoolResizeAbort(storagePoolUUID string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}
	// Reset the new size in spec. The storagePool error conditions can then be cleared by Podrick.
	storagePool.Spec.Size = storagePool.Status.Size
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	return storagePool, err
}

// Note that if the migration has completed by the time this operation is
// issued, the operation will do nothing
func (client *NVCClient) NetAppStoragePoolMigrateZoneAbort(
	storagePoolUUID string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}
	if storagePool.Spec.Zone == storagePool.Status.Zone {
		return nil, MigrateZoneAbortNoMigrationFailureError{}
	}

	if storagePool.Spec.LastZoneMigrationCause == quarkconfig.ZoneMigrationCauseFailure {
		// return an error here, since we can't abort zone migrations due to failure
		return nil, MigrateZoneAbortZoneFailureError{}
	}
	storagePool.Spec.Zone = storagePool.Status.Zone
	storagePool.Spec.LastZoneMigrationCause = storagePool.Status.LastZoneMigrationCause
	storagePool.Spec.ZoneMigrationCount = storagePool.Status.ZoneMigrationCount
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	return storagePool, err
}

// NetAppStoragePoolResized returns "true" when the data PVCs are correctly resized.
func (client *NVCClient) NetAppStoragePoolResized(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return sp.Status.Size.Cmp(sp.Spec.Size) == 0
}

func (client *NVCClient) NetAppStoragePoolMigratedZone(sp *quarkv1alpha1.NetAppStoragePool) bool {
	// Only acknowledge a zone migration if it wasn't initiated as part of a zone
	// failure.
	return sp.Status.Zone == sp.Spec.Zone && sp.Spec.LastZoneMigrationCause == quarkconfig.ZoneMigrationCauseUser
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdated(
	sp *quarkv1alpha1.NetAppStoragePool,
) bool {
	// If global access enablement is being aborted, the spec and status fields
	// will be equal, but GlobalAccessAborting will be true. We only want to
	// declare this to be finalized once GlobalAccessAborting is set to false.
	return sp.Status.GlobalAccess == sp.Spec.GlobalAccess &&
		!sp.Spec.GlobalAccessAborting
}

// NetAppStoragePoolADDNSUpdated returns "true" when AD DNSIPs are applied to the storagePool.
func (client *NVCClient) NetAppStoragePoolADDNSUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsEqualStrings(sp.Spec.ADDNSIPs, sp.Status.ADDNSIPs)
}

// NetAppStoragePoolADPasswordResetDone returns "true" when machine-account password reset is successful.
func (client *NVCClient) NetAppStoragePoolADPasswordResetDone(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return (sp.Status.SMBPasswordResetCount == sp.Spec.SMBPasswordResetCount)
}

// NetAppStoragePoolSMBADConfigurationUpdated returns "true" when AD configurations updated for the storagePool.
func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return (sp.Status.SMBADConfiguration == sp.Spec.SMBADConfiguration)
}

// NetAppStoragePoolServerConfigurationUpdated returns "true" when SMB server configurations are updated for the storagePool.
func (client *NVCClient) NetAppStoragePoolSMBServerConfigurationUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsSMBServerConfigurationEqual(sp)
}

// NetAppStoragePoolSMBADUserListUpdated returns "true" when all users are updated for a given storagePool.
func (client *NVCClient) NetAppStoragePoolSMBADUserListUpdated(sp *quarkv1alpha1.NetAppStoragePool) (ret bool) {
	if !client.EnableCCPd || !client.EnableBackupSecurityUsers {
		return false
	}
	/*
	 To maintain backward compatability, isCounterEqual is set to true
	 when sp.Status.SMBUsersUpdateCount is 0
	*/
	isCounterEqual := sp.Status.SMBUsersUpdateCount == 0 || sp.Spec.SMBUsersUpdateCount == sp.Status.SMBUsersUpdateCount
	val := (util.IsEqualStrings(sp.Spec.SMBBackupUsers, sp.Status.SMBBackupUsers) &&
		util.IsEqualStrings(sp.Spec.SMBSecurityPrivilegeUsers, sp.Status.SMBSecurityPrivilegeUsers) && isCounterEqual)

	return val
}

// NetAppStoragePoolUpgraded returns "true" when StoragePool version upgrade is successful.
func (client *NVCClient) NetAppStoragePoolUpgraded(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	conditions := storagePool.Status.Conditions
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltSucceeded || cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltRequested || cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeFailed {
			// return false if another upgrade is in progress
			return false
		}
	}
	return (storagePool.Status.QuarkVersion == storagePool.Spec.QuarkVersion)

}

// NetAppStoragePoolDelete is used to signal the deletion of a NetAppStoragePool.
type NetAppStoragePoolDelete struct {
	StoragePool *quarkv1alpha1.NetAppStoragePool
}

// NetAppStoragePoolUpdate is used to signal the most recent state of a NetAppStoragePool and the
// most recent changes.
type NetAppStoragePoolUpdate struct {
	StoragePool            *quarkv1alpha1.NetAppStoragePool
	Status                 *NetAppStoragePoolUpdateStatus
	UpdatedConditions      []quarkv1alpha1.NetAppStoragePoolCondition
	LastZoneMigrationCause string
	Error                  error
}

// NetAppStoragePoolUpdateStatus captures the status of any pending updates.
type NetAppStoragePoolUpdateStatus struct {
	SizeUpdated                    bool
	ZoneUpdated                    bool
	GlobalAccessUpdated            bool
	ADDNSUpdated                   bool
	SMBADConfigurationUpdated      bool
	SMBADConfigUserListUpdated     bool
	SMBServerConfigurationUpdated  bool
	ADPasswordReset                bool
	QuarkVersionUpdated            bool
	LdapConfigured                 bool
	StoragePoolDisabledUpdated     bool
	KerberosConfigured             bool
	KerberosConfigureCreateAborted bool
	EncryptionInitialized          bool
}

// NetAppStoragePoolUpdated returns "true" when the actual state has converged to the
// desired state. If not, it will return an object describing what actions are pending.
func (client *NVCClient) NetAppStoragePoolUpdated(storagePool *quarkv1alpha1.NetAppStoragePool) (bool, *NetAppStoragePoolUpdateStatus, error) {
	statusConverged := true

	status := &NetAppStoragePoolUpdateStatus{
		SizeUpdated:                    true,
		ZoneUpdated:                    true,
		GlobalAccessUpdated:            true,
		ADDNSUpdated:                   true,
		SMBADConfigurationUpdated:      true,
		SMBADConfigUserListUpdated:     true,
		SMBServerConfigurationUpdated:  true,
		ADPasswordReset:                true,
		QuarkVersionUpdated:            true,
		LdapConfigured:                 true,
		StoragePoolDisabledUpdated:     true,
		KerberosConfigured:             true,
		KerberosConfigureCreateAborted: true,
		EncryptionInitialized:          true,
	}
	// Check size
	if !client.NetAppStoragePoolResized(storagePool) {
		status.SizeUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolMigratedZone(storagePool) {
		status.ZoneUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolGlobalAccessUpdated(storagePool) {
		status.GlobalAccessUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolADDNSUpdated(storagePool) {
		status.ADDNSUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBADConfigurationUpdated(storagePool) {
		status.SMBADConfigurationUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBADUserListUpdated(storagePool) {
		status.SMBADConfigUserListUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolLdapReady(storagePool) {
		status.LdapConfigured = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBServerConfigurationUpdated(storagePool) {
		status.SMBServerConfigurationUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolADPasswordResetDone(storagePool) {
		status.ADPasswordReset = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolUpgraded(storagePool) {
		status.QuarkVersionUpdated = false
		statusConverged = false
	}
	if storagePool.Spec.Disabled != storagePool.Status.Disabled {
		status.StoragePoolDisabledUpdated = false
		statusConverged = false
	}
	if kerbReady, _ := client.NetAppStoragePoolKerberosReady(storagePool); !kerbReady {
		status.KerberosConfigured = false
		statusConverged = false
	}
	if kerbConfigAborted, _ := client.NetAppStoragePoolKerberosAborted(storagePool); !kerbConfigAborted {
		status.KerberosConfigureCreateAborted = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolEncryptionInitialization(storagePool) {
		status.EncryptionInitialized = false
		statusConverged = false
	}

	// Return the status of any potentially pending updates.
	if statusConverged {
		return true, status, nil
	} else {
		return false, status, nil
	}
}

func (client *NVCClient) handleStoragePoolDeletes(sp *quarkv1alpha1.NetAppStoragePool) {
	client.storagePoolDeleteChannel <- &NetAppStoragePoolDelete{sp}
}

func (client *NVCClient) handleStoragePoolUpdates(oldSp *quarkv1alpha1.NetAppStoragePool, newSp *quarkv1alpha1.NetAppStoragePool) {
	sendEvent := false

	// We send events when any of the NetAppStoragePool has converged to the desired state.
	_, newStatus, _ := client.NetAppStoragePoolUpdated(newSp)
	_, oldStatus, _ := client.NetAppStoragePoolUpdated(oldSp)
	if newStatus != oldStatus {
		log.WithFields(log.Fields{
			"storagePool": newSp.Name,
			"newStatus":   newStatus,
		}).Debug("NetApp Storage pool state has been updated.")
		sendEvent = true
	}
	// We send events when the NetStoragePool's conditions have changed.
	if len(newSp.Status.Conditions) != len(oldSp.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"storagePool": newSp.Name,
			"conditions":  newSp.Status.Conditions,
		}).Debug("NetApp Storage pool has updated conditions.")
	} else {
		for i, cond := range newSp.Status.Conditions {
			if oldSp.Status.Conditions[i].Type != cond.Type || oldSp.Status.Conditions[i].Status != cond.Status {
				sendEvent = true
				log.WithFields(log.Fields{
					"storagePool": newSp.Name,
					"conditions":  newSp.Status.Conditions,
				}).Debug("NetApp Storagepool has updated conditions.")
				break
			}
		}
	}
	if newSp.Spec.LastZoneMigrationCause != oldSp.Spec.LastZoneMigrationCause {
		sendEvent = true
	}
	// If the address exhaustion condition is set we need to check if Observed Subnets in Status is
	// not the same as Spec. We need to check in event handler if the address exhaustion error needs
	// to be propagated to SDE as a fatal error.
	if (util.GetNetAppStoragePoolCondition(newSp, quarkv1alpha1.NetAppStoragePoolAddressExhaustionError) != nil) &&
		util.IsEqualStrings(newSp.Status.ObservedILBSubnets, newSp.Spec.IlbSubnets) {
		sendEvent = true
	}

	if sendEvent {
		client.storagePoolUpdateChannel <- &NetAppStoragePoolUpdate{
			StoragePool:            newSp,
			Status:                 newStatus,
			UpdatedConditions:      newSp.Status.Conditions,
			LastZoneMigrationCause: newSp.Spec.LastZoneMigrationCause,
		}
	}
}

func (client *NVCClient) validateStoragePoolSize(size resource.Quantity) (err error) {
	var (
		maxStoragePoolSizeAllowed resource.Quantity
	)
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return fmt.Errorf("no quarks found")
	}
	quark := quarks[0]
	if quark.Spec.Cloud == quarkconfig.CloudAzure {
		maxStoragePoolSizeAllowed = maxStoragePoolSizeAzure
	} else {
		maxStoragePoolSizeAllowed = maxStoragePoolSize
	}
	minStoragePoolSizeDifference := size.Value() - minStoragePoolSize.Value()
	maxStoragePoolSizeDifference := size.Value() - maxStoragePoolSizeAllowed.Value()
	if minStoragePoolSizeDifference < 0 || maxStoragePoolSizeDifference > 0 || minStoragePoolSizeDifference%incrementalSize.Value() != 0 {
		return fmt.Errorf("size=%s, storage pool size should be a minimum of %s, maximum of %s and a multiple of %s", size.String(), minStoragePoolSize.String(), maxStoragePoolSizeAllowed.String(), incrementalSize.String())
	}
	return nil
}

func (client *NVCClient) NetAppStoragePoolSizeUpdate(sp *quarkv1alpha1.NetAppStoragePool, newSizeString string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	size, err := resource.ParseQuantity(newSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}

	err = client.validateStoragePoolSize(size)
	if err != nil {
		return nil, err
	}

	if !client.EnableGcp1p {
		if sp.Spec.Size.Value() > size.Value() {
			return nil, fmt.Errorf("storage pool size shrink operation is not allowed. Current storagePool size is %s, newsize=%s", sp.Spec.Size.String(), size.String())
		}
	}
	sp.Spec.Size = size
	sp, err = client.netAppStoragePoolUpdate(sp)

	return sp, err
}

func (client *NVCClient) NetAppStoragePoolMigrateZone(
	sp *quarkv1alpha1.NetAppStoragePool, zone string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if sp.Spec.Zone == zone {
		// We already did the migration; don't update.
		return sp, nil
	}
	// Validate zone choice
	zoneOk := false
	if len(sp.Status.AvailableZones) == 0 {
		return nil, fmt.Errorf("unable to migrate; no AvailableZones to choose from")
	}

	for i := 0; i < len(sp.Status.AvailableZones) && !zoneOk; i++ {
		zoneOk = zone == sp.Status.AvailableZones[i]
	}
	if !zoneOk {
		return nil, fmt.Errorf("zone %s is not in AvailableZones, choose one of %v", zone, sp.Status.AvailableZones)
	}

	sp.Spec.Zone = zone
	sp.Spec.LastZoneMigrationCause = quarkconfig.ZoneMigrationCauseUser
	sp.Spec.ZoneMigrationCount = sp.Spec.ZoneMigrationCount + 1
	// This is subject to clock skew, but there's not a good way around this.
	now := metav1.Now()
	sp.Spec.LastZoneMigrationTime = &now
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdate(
	sp *quarkv1alpha1.NetAppStoragePool, enabled bool,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if sp.Spec.GlobalAccessAborting {
		return nil, GlobalAccessAbortInProgressError{}
	}
	sp.Spec.GlobalAccess = enabled
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessSet(
	sp *quarkv1alpha1.NetAppStoragePool, enabled bool,
) *quarkv1alpha1.NetAppStoragePool {
	sp.Spec.GlobalAccess = enabled
	return sp
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdateAbort(
	storagePoolUUID string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	sp, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}
	if sp.Status.GlobalAccess == sp.Spec.GlobalAccess {
		return nil, GlobalAccessUpdateAbortNoUpdateError{}
	}
	sp.Spec.GlobalAccess = sp.Status.GlobalAccess
	if nvc.HasFinalizer(sp.Finalizers,
		quarkconfig.ADClientAddressFinalizer) && sp.Status.ADClientIP == "" {
		// We're in the process of first time enablement; mark that we're
		// aborting
		sp.Spec.GlobalAccessAborting = true
	}
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) netAppStoragePoolStatusUpdate(storagePool *quarkv1alpha1.NetAppStoragePool) (*quarkv1alpha1.NetAppStoragePool, error) {
	result, err := client.updateStatusWithRetry(storagePool)
	if err != nil {
		log.WithFields(log.Fields{
			"storagepool": storagePool.Name,
			"error":       err,
		}).Error("updating status failed.")

		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}

		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppStoragePool), nil
}

func (client *NVCClient) netappStoragePoolConditionUpdate(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	abortOperation bool,
	conditionTypes ...quarkv1alpha1.NetAppStoragePoolConditionType,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	updateStatus := false

	for _, conditionType := range conditionTypes {
		conditionPresent, retryErr := util.IsErrorConditionPresent(storagePool.Status.Conditions,
			conditionType)
		if conditionPresent {
			if abortOperation {
				storagePool.Status.Conditions = util.RemoveStoragePoolConditions(storagePool.Status.Conditions,
					conditionType)
				updateStatus = true
			} else if !retryErr {
				// Setting the ForceRetry flag in condition so that podrick can treat this condition as a retry error
				storagePool.Status.Conditions = util.SetForceRetryInErrorCondition(storagePool.Status.Conditions,
					conditionType)
				updateStatus = true
			}
		}
	}

	if updateStatus {
		return client.netAppStoragePoolStatusUpdate(storagePool)
	}
	return storagePool, nil
}

// Reset Machine Account password for SMB storagePool
func (client *NVCClient) NetAppStoragePoolSMBPasswordReset(storagePoolUUID string, user, password string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}
	secretName := util.GetVolPwdResetADSecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": storagePool.Name,
			"secret": secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	storagePool.Spec.SMBPasswordResetCount = storagePool.Spec.SMBPasswordResetCount + 1
	newStoragePool, err := client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		delErr := client.SecretDelete(secretName)
		if delErr != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"secret":      secretName,
				"error":       delErr,
			}).Error("Failed to delete secret.")
		}

		return nil, err
	}
	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		newStoragePool, err = client.netappStoragePoolConditionUpdate(newStoragePool, false,
			quarkv1alpha1.NetAppStoragePoolSMBPasswordResetError)
	}

	return newStoragePool, err
}

// Abort machine-account password reset operation.
// SDE should abort the password reset operation if it times out.
// Abort will delete the secret and avoid retrying operation.
func (client *NVCClient) NetAppStoragePoolADPasswordResetAbort(storagePoolUUID string) (sp *quarkv1alpha1.NetAppStoragePool, err error) {
	secretName := util.GetVolPwdResetADSecretName(storagePoolUUID)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	sp, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}

	sp.Spec.SMBPasswordResetCount = sp.Status.SMBPasswordResetCount
	sp, err = client.netAppStoragePoolUpdate(sp)

	// Cleanup up the error conditions
	if err == nil {
		sp, err = client.netappStoragePoolConditionUpdate(sp, true,
			quarkv1alpha1.NetAppStoragePoolSMBPasswordResetError)
	}

	return sp, err
}

// Update DNS IP for AD storagePool
func (client *NVCClient) NetAppStoragePoolADDNSUpdate(sp *quarkv1alpha1.NetAppStoragePool, dnsIPs []string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	// Returns error if the given IP address is not valid IPv4 address
	err = util.ValidateIPv4(dnsIPs)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.ADDNSIPs = dnsIPs
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}

	return
}

func (client *NVCClient) NetAppStoragePoolADDNSUpdateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.ADDNSIPs = storagePool.Status.ADDNSIPs
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleanup up the error conditions
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}

	return
}

// Update EncryptedCommunication for AD storagePool
func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdate(sp *quarkv1alpha1.NetAppStoragePool, SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.SMBADConfiguration = SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}
	return
}

// Update backup and security users from AD settings
func (client *NVCClient) NetAppStoragePoolSMBADBackupSecurityUsersUpdate(sp *quarkv1alpha1.NetAppStoragePool, backup_user_list []string, security_privilege_user_list []string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	if !client.EnableCCPd || !client.EnableBackupSecurityUsers {
		return nil, fmt.Errorf("Cannot update users as CCPd flag or Backup/Security feature flag is disabled")
	}
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	if client.NetAppStoragePoolSMBReady(storagePool) == false {
		return nil, fmt.Errorf("Cannot update users as SMB config is not ready")
	}

	err = client.validateNoDuplicates(backup_user_list, security_privilege_user_list)
	if err != nil {
		return nil, err
	}
	err = client.validateNames(backup_user_list, security_privilege_user_list)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.SMBBackupUsers = backup_user_list
	storagePool.Spec.SMBSecurityPrivilegeUsers = security_privilege_user_list
	storagePool.Spec.SMBUsersUpdateCount = storagePool.Spec.SMBUsersUpdateCount + 1
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleanup up the error conditions
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolSMBBackupUsersError, quarkv1alpha1.NetAppStoragePoolSMBSecurityPrivilegeUsersError)
	}

	return storagePool, err
}

func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.SMBADConfiguration = storagePool.Status.SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleanup up the error conditions
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}

	return
}

func (client *NVCClient) NetAppStoragePoolSMBConfigCreate(sp *quarkv1alpha1.NetAppStoragePool, user, password string, ADDNSIPs []string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration, SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}

	// Returns error if the given IP address is not valid IPv4 address
	err = util.ValidateIPv4(ADDNSIPs)
	if err != nil {
		return nil, err
	}

	if client.NetAppStoragePoolSMBInitialized(storagePool) {
		if storagePool.Status.SMBServerConfiguration != SMBServerConfiguration || storagePool.Status.SMBADConfiguration != SMBADConfiguration ||
			!util.IsEqualStrings(storagePool.Status.ADDNSIPs, ADDNSIPs) {
			return nil, fmt.Errorf("updating CIFS Server config is not allowed: Server already configured")
		}
	}

	secretName := util.GetVolCreateADSecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool.Spec.ADDNSIPs = ADDNSIPs
	storagePool.Spec.SMBADConfiguration = SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError,
			quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
	}
	return
}

func (client *NVCClient) NetAppStoragePoolSMBConfigCreateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	secretName := util.GetVolCreateADSecretName(sp.Name)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.ADDNSIPs = storagePool.Status.ADDNSIPs
	storagePool.Spec.SMBServerConfiguration = storagePool.Status.SMBServerConfiguration
	storagePool.Spec.SMBADConfiguration = storagePool.Status.SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleanup up the error conditions
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError,
			quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
	}

	return
}

func (client *NVCClient) NetAppStoragePoolSMBServerConfigurationUpdate(sp *quarkv1alpha1.NetAppStoragePool, user, password string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	if storagePool.Status.SMBServerConfiguration.Netbios == "" {
		return nil, fmt.Errorf("updating CIFS Server config is not allowed: Server not yet configured")
	}

	if SMBServerConfiguration.Netbios == "" || SMBServerConfiguration.Domain == "" {
		return nil, fmt.Errorf("Netbios and Domain cannot be empty.")
	}

	// If IsSMBServerConfigurationEqual is false then previous SMB config modify is failed or not completed.
	// In that case allow to continue modify the username and password
	if util.IsSMBServerConfigurationEqual(storagePool) &&
		SMBServerConfiguration.Netbios == storagePool.Spec.SMBServerConfiguration.Netbios &&
		SMBServerConfiguration.OrganizationUnit == storagePool.Spec.SMBServerConfiguration.OrganizationUnit &&
		SMBServerConfiguration.Domain == storagePool.Spec.SMBServerConfiguration.Domain &&
		SMBServerConfiguration.PreferredADSite == storagePool.Spec.SMBServerConfiguration.PreferredADSite {
		return nil, fmt.Errorf("Invalid inputs. OrganizationUnit, NetBIOS, Domain and Preferred AD site are same as current configuration.")
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}

	secretName := util.GetSMBServerModifySecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	// Increment counters if Domain is modified. This will trigger workflow in CCPd.
	if SMBServerConfiguration.Domain != storagePool.Spec.SMBServerConfiguration.Domain {
		if storagePool.Spec.LdapEnabled {
			storagePool.Spec.LDAPUpdateCount += 1
		}
		if client.EnableBackupSecurityUsers {
			storagePool.Spec.SMBUsersUpdateCount += 1
		}
	}
	// Increment counters if Site is modified. This will trigger workflow in CCPd.
	if SMBServerConfiguration.PreferredADSite != storagePool.Spec.SMBServerConfiguration.PreferredADSite {
		if storagePool.Spec.LdapEnabled {
			storagePool.Spec.LDAPUpdateCount += 1
		}
	}

	// updating kerberos server configuration if required
	if sp.Spec.KerbServerConfiguration != nil && storagePool.Spec.KerbServerConfiguration != nil &&
		sp.Spec.KerbServerConfiguration != storagePool.Spec.KerbServerConfiguration {
		storagePool.Spec.KerbServerConfiguration = sp.Spec.KerbServerConfiguration
	}
	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
	}

	return
}

// Validate availableZones, verify zone exists in availableZones, validate zone is non-empty
// for zonal storage pools in regional clusters.
func (client *NVCClient) validateZones(serviceLevel ServiceLevel, zone string, availableZones []string) (bool, error) {
	var err error

	switch serviceLevel {
	// For zonal SPs in regional cluster, only one availability zone should be passed.
	case ServiceLevelStandard:
		if zone == "" {
			return false, fmt.Errorf("specify target zone when deploying zonal storagePool in a regional cluster")
		}
		if len(availableZones) == quarkconfig.AvailableZonesGCP1P && zone == availableZones[0] {
			return true, nil
		} else if len(availableZones) != quarkconfig.AvailableZonesGCP1P {
			return false, fmt.Errorf("invalid number of availableZones, "+
				"expected %d, received %d",
				quarkconfig.AvailableZonesGCP1P, len(availableZones))
		} else if zone != availableZones[0] {
			return false, fmt.Errorf("target zone doesn't exist in availableZones, "+
				"expected %s, received %s",
				availableZones[0], zone)
		}
	case ServiceLevelRegionalStandard:
		if len(availableZones) == 0 {
			// We can use the default behavior; no reason to look up the cloud
			// provider
			return true, nil
		}
		// Need to know what cloud provider
		quarks, err := client.NetAppQuarkGetAll()
		if err != nil {
			return false, fmt.Errorf("unable to get Cloud provider, error retrieving quark CR: %v", err)
		}
		if len(quarks) == 0 {
			return false, fmt.Errorf("unable to get Cloud provider, quark CR not found")
		}
		quark := quarks[0]
		if quark.Spec.Cloud == quarkconfig.CloudGCP {
			if len(availableZones) != quarkconfig.AvailableZonesGCP {
				return false, fmt.Errorf("invalid number of availableZones, "+
					"for Cloud provider, %s, expect %d, received %d",
					quark.Spec.Cloud, quarkconfig.AvailableZonesGCP, len(availableZones))
			}
			if !strings.Contains(strings.Join(availableZones, ","), zone) {
				return false, fmt.Errorf("target zone doesn't exist in availableZones, "+
					"expected %s, received %s",
					availableZones[0], zone)
			}
		} else if quark.Spec.Cloud == quarkconfig.CloudAzure {
			if len(availableZones) != quarkconfig.AvailableZonesAzure {
				return false, fmt.Errorf("invalid number of availableZones, "+
					"for Cloud provider, %s, expect %d, received %d",
					quark.Spec.Cloud, quarkconfig.AvailableZonesAzure, len(availableZones))
			}
		}
	}
	return true, err
}

// ConstructRegionalNetAppStoragePoolWithSMBConfig() does validations on availableZones before calling
// ConstructNetAppStoragePoolWithSMBConfig.  It then sets Zone and AvailableZones fields of storagePool.Spec
func (client *NVCClient) ConstructRegionalNetAppStoragePoolWithSMBConfig(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	zone string,
	availableZones []string,
	serviceLevel ServiceLevel,
	ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	zonesOk, err := client.validateZones(serviceLevel, zone, availableZones)
	if !zonesOk {
		return nil, err
	}

	storagePool, err = client.ConstructNetAppStoragePoolWithSMBConfig(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel, ADDNSIPs, SMBServerConfiguration, SMBADConfiguration)
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}

	// Set Zone and availableZones
	storagePool.Spec.Zone = zone
	storagePool.Spec.AvailableZones = availableZones

	return storagePool, err
}

func (client *NVCClient) validateNoDuplicates(backup_user_list []string, security_privilege_user_list []string) (err error) {
	BackupUserVisitMap := make(map[string]bool, 0)
	for i := 0; i < len(backup_user_list); i++ {
		if BackupUserVisitMap[backup_user_list[i]] == true {
			return fmt.Errorf("Duplicate name %s present in Backup user list", backup_user_list[i])
		} else {
			BackupUserVisitMap[backup_user_list[i]] = true
		}
	}

	SecurityPrivilegeUserVisitMap := make(map[string]bool, 0)
	for i := 0; i < len(security_privilege_user_list); i++ {
		if SecurityPrivilegeUserVisitMap[security_privilege_user_list[i]] == true {
			return fmt.Errorf("Duplicate name %s present in Security privilege user list", security_privilege_user_list[i])
		} else {
			SecurityPrivilegeUserVisitMap[security_privilege_user_list[i]] = true
		}
	}

	return nil
}

func (client *NVCClient) validateNames(backup_user_list []string, security_privilege_user_list []string) (err error) {

	// Validation reference: Cloud Volumes Active directory name validations for backup and security privilege users.
	var max_len = 255
	invalid_chars := "\"/[]:;|=+*?<>@"
	for _, name := range backup_user_list {
		if strings.ContainsAny(name, invalid_chars) {
			return fmt.Errorf("Invalid characters in backup user %s, cannot contain any of %s", name, invalid_chars)
		}
		if len(name) > max_len {
			return fmt.Errorf("Length of name %s cannot be larger than %d", name, max_len)
		}
	}
	for _, name := range security_privilege_user_list {
		if strings.ContainsAny(name, invalid_chars) {
			return fmt.Errorf("Invalid characters in security privilege user %s, cannot contain any of %s", name, invalid_chars)
		}
		if len(name) > max_len {
			return fmt.Errorf("Length of name %s cannot be larger than %d", name, max_len)
		}
	}
	return nil
}

func (client *NVCClient) ValidateLdapConfig(ADDNSIPs []string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration) (err error) {
	ADDNSIPsString := strings.Join(ADDNSIPs, ",")
	if ADDNSIPsString == "" || SMBServerConfiguration.Domain == "" ||
		SMBServerConfiguration.Netbios == "" {
		return fmt.Errorf("cannot create LDAP configuration: one or more empty SMB configuration parameters")
	}
	return nil
}

func (client *NVCClient) ConstructNetAppStoragePoolWithSMBConfig(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	serviceLevel ServiceLevel,
	ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.ConstructNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel)
	if storagePool == nil || err != nil {
		log.Fatalf("Failed to construct StoragePool")
		return
	}

	err = util.ValidateIPv4(ADDNSIPs)
	if err != nil {
		return nil, err
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool.Spec.ADDNSIPs = ADDNSIPs
	storagePool.Spec.SMBADConfiguration = SMBADConfiguration

	return storagePool, err
}

// calls ConstructRegionalNetAppStoragePoolWithSMBConfig() and sets LDAPEnabled to true in storagePool.Spec
func (client *NVCClient) ConstructNetAppStoragePoolWithSMBConfigAndLDAP(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	zone string,
	availableZones []string,
	serviceLevel ServiceLevel,
	ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	if !client.EnableCCPd || !client.EnableLdap {
		return nil, fmt.Errorf("cannot create LDAP configuration: CCPd flag or LDAP feature flag is disabled")
	}
	err = client.ValidateLdapConfig(ADDNSIPs, SMBServerConfiguration)
	if err != nil {
		return nil, err
	}

	// For regional storage pools (or) 1P config, we use 'ConstructRegionalNetAppStoragePool'
	if serviceLevel == ServiceLevelRegionalStandard || client.EnableGcp1p {
		storagePool, err = client.ConstructRegionalNetAppStoragePoolWithSMBConfig(ownerID,
			displayName, sizeString, storagePoolUUID, zone, availableZones, serviceLevel,
			ADDNSIPs, SMBServerConfiguration, SMBADConfiguration)
	} else {
		storagePool, err = client.ConstructNetAppStoragePoolWithSMBConfig(ownerID,
			displayName, sizeString, storagePoolUUID, serviceLevel, ADDNSIPs,
			SMBServerConfiguration, SMBADConfiguration)
	}
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}
	storagePool.Spec.LdapEnabled = true

	return storagePool, err
}

// ConstructRegionalNetAppStoragePool() does validations on availableZones before calling
// ConstructNetAppStoragePool.  It then sets Zone and AvailableZones fields of storagePool.Spec
func (client *NVCClient) ConstructRegionalNetAppStoragePool(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	zone string,
	availableZones []string,
	serviceLevel ServiceLevel,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	zonesOk, err := client.validateZones(serviceLevel, zone, availableZones)
	if !zonesOk {
		return nil, err
	}

	storagePool, err = client.ConstructNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel)
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}

	// Set Zone and availableZones
	storagePool.Spec.Zone = zone
	storagePool.Spec.AvailableZones = availableZones

	return storagePool, err
}

// Constructs NetAppStoragePool definition.
// ownerID: OwnerID of the StoragePool.
// dispalyName: Name of the StoragePool.
// sizeString: StoragePool Size.
// storagePoolUUID: Unique ID of storagePool in k8s cluster.
// serviceLevel: Standard for alpha.
func (client *NVCClient) ConstructNetAppStoragePool(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	serviceLevel ServiceLevel,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	defer func() {
		if e := recover(); e != nil {
			err = fmt.Errorf("invalid value for size")
		}
	}()

	// Validate service level
	switch serviceLevel {
	case ServiceLevelStandard, ServiceLevelRegionalStandard:
	default:
		return nil, fmt.Errorf("invalid value for service level")
	}

	// validate UUID
	_, err = uuid.FromString(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for UUID")
	}
	// Validate size
	size, err := resource.ParseQuantity(sizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	err = client.validateStoragePoolSize(size)
	if err != nil {
		return nil, err
	}

	// Validate ownerID
	if ownerID == "" {
		return nil, fmt.Errorf("storage pool OwnerID required")
	}

	storagePool = &quarkv1alpha1.NetAppStoragePool{
		metav1.TypeMeta{},
		metav1.ObjectMeta{
			Name: storagePoolUUID,
			Labels: map[string]string{
				quarkconfig.StoragePoolLabel: storagePoolUUID,
				quarkconfig.OwnerIDLabel:     ownerID,
			},
		},
		quarkv1alpha1.NetAppStoragePoolSpec{
			Size:         size,
			DisplayName:  displayName,
			ServiceLevel: string(serviceLevel),
		},
		quarkv1alpha1.NetAppStoragePoolStatus{},
	}
	return storagePool, err
}

// Label a quark storage pool with SDE's storage pool uuid.
// This function expects 2 arguments .
// storagePoolUUID  is the name of the sp to label. sdeSpUUID is SP uuid assigned by SDE.
func (client *NVCClient) LabelNetAppStoragePool(storagePoolUUID, sdeSpUUID string) error {
	if sdeSpUUID == "" {
		return fmt.Errorf("provide non empty uuid of storage pool")
	}

	sp, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.Background(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return err
	}
	if len(sp.Labels) == 0 {
		sp.Labels = make(map[string]string)
	}

	sp.Labels[quarkconfig.SDESpUUID] = sdeSpUUID

	_, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Update(context.Background(), sp, metav1.UpdateOptions{})

	return err
}

func updateCMEKConfigMapOwnerRef(client *NVCClient,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	cmekConfigMap *corev1.ConfigMap,
) (err error) {

	cmekConfigMap.OwnerReferences = append(cmekConfigMap.OwnerReferences,
		*metav1.NewControllerRef(
			storagePool,
			schema.GroupVersionKind{
				Group:   quarkv1alpha1.SchemeGroupVersion.Group,
				Version: quarkv1alpha1.SchemeGroupVersion.Version,
				Kind:    "NetAppStoragePool",
			},
		),
	)
	_, err = client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Update(context.TODO(), cmekConfigMap, metav1.UpdateOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"configMap":   cmekConfigMap.Name,
		}).Error("Failed to add owner reference to cmek config map: ", err)
		return err
	}

	return nil
}

// optionsFn: Additional config for new features like cmek etc - clients can provide optional additional config by providing a
//
//	         variable number of options functions. An example option function is:
//	         func WithCmekConfig(cmekCfg *quarkv1alpha1.CMEKConfiguration) quarkv1alpha1.StoragePoolOptionsFn {
//					return func(additionalSPCfg *quarkv1alpha1.AdditionalStoragePoolConfig) {
//								additionalSPCfg.CMEKConfig = cmekCfg
//						   }
//	         }
func (client *NVCClient) NetAppStoragePoolCreate(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	user,
	password string,
	optionFns ...quarkv1alpha1.StoragePoolOptionsFn,
) (sp *quarkv1alpha1.NetAppStoragePool, err error) {

	var secretName string
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("no quarks found")
	}
	quark := quarks[0]
	// add quark version to create NetAppStoragePool only for "Production" & "TestVariant" configs
	switch strings.ToLower(quark.Spec.ControllerConfig.Variant) {
	case quarkconfig.ProductionVariant, quarkconfig.OnPremVariant:
		storagePool.Spec.QuarkVersion = quark.Status.Version
	}
	if storagePool.Spec.SMBServerConfiguration.Netbios != "" {
		secretName = util.GetVolCreateADSecretName(storagePool.Name)
		_, err = client.createOrUpdateADSecret(secretName, user, password)
		if err != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"secret":      secretName,
			}).Error("Failed to create secret: ", err)
			return nil, err
		}
	}

	additionalConfig := quarkv1alpha1.AdditionalStoragePoolConfig{}
	for _, optionFn := range optionFns {
		optionFn(&additionalConfig)
	}

	if quark.Spec.EnableGcp1p {
		if len(additionalConfig.IlbSubnets) == 0 {
			return nil, fmt.Errorf("failed to create Storage pool: no subnets are specified")
		}
		storagePool.Spec.IlbSubnets = additionalConfig.IlbSubnets
	}

	var cmekConfigMap *corev1.ConfigMap
	cmekEnabled := quark.Spec.EnableCMEK && (additionalConfig.CMEKConfig != nil)
	// TODO-cmek add gcp1p flag whenever cmek feature is lightened
	if cmekEnabled {
		// Check if a Storage Pool with the same UUID already exists. If it does do not create CMEK configmap as eventually the
		// Storage Pool would not be created as well.
		_, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.TODO(), storagePool.Name, metav1.GetOptions{})
		if err == nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
			}).Debug("storage pool with this name already exists", err)
			return nil, fmt.Errorf("storage pool with this name already exists")
		}

		// If the CMEKEmulatorConfig exists we need to set couple of additional fields in the CMEK ConfigMap for Emulator Testing purpose
		if cmekConfigMap, err = client.createCmekConfigMap(storagePool.Name, additionalConfig.CMEKConfig, additionalConfig.CMEKEmulatorConfig); err != nil {
			return nil, err
		}
		storagePool.Labels[quarkconfig.CMEKConfigMapLabel] = cmekConfigMap.Name
	}

	result, err := client.createWithRetry(storagePool)
	if err != nil {
		// If failed to create Storage Pool and CMEK is enabled, delete the CMEK ConfigMap
		if cmekEnabled {
			err := client.deleteCMEKConfigMap(cmekConfigMap.Name)
			if err != nil {
				log.Errorf("Failed to delete the CMEK ConfigMap created while creating storage pool: %v", err)
			}
		}
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create storagePool:", err)
		return nil, err
	}

	sp = result.(*quarkv1alpha1.NetAppStoragePool)

	// Update the CMEK ConfigMap owner reference
	if cmekEnabled {
		err = updateCMEKConfigMapOwnerRef(client, sp, cmekConfigMap)
		if err != nil {
			return nil, err
		}
	}

	return sp, nil
}

func (client *NVCClient) createCmekConfigMap(
	storagePoolName string,
	cmekConfig *quarkv1alpha1.CMEKConfiguration,
	cmekEmulatorConfig *quarkv1alpha1.CMEKEmulatorConfiguration,
) (*corev1.ConfigMap, error) {
	// Check all needed CMEK configs are present
	var errMsg string
	if cmekConfig == nil {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Cmek configs")
	} else if strings.ToLower(cmekConfig.Cloud) != "gcp" {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Cloud=gcp")
	} else if len(strings.TrimSpace(cmekConfig.ProjectID)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Project Id")
	} else if len(strings.TrimSpace(cmekConfig.KeyRing)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Key Ring")
	} else if len(strings.TrimSpace(cmekConfig.KeyRingLocation)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Key Ring Location")
	} else if len(strings.TrimSpace(cmekConfig.KeyName)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Key Name")
	}
	if len(strings.TrimSpace(errMsg)) != 0 {
		log.WithFields(log.Fields{
			"storagePool": storagePoolName,
		}).Error(errMsg)
		return nil, fmt.Errorf(errMsg)
	}

	cmekConfigMapName := quarkconfig.CMEKConfigMapPrefix + uuid.NewV1().String()

	cmekConfigMap := &corev1.ConfigMap{
		ObjectMeta: metav1.ObjectMeta{
			Name:      cmekConfigMapName,
			Namespace: client.nvcNamespace,
		},
		Data: map[string]string{
			quarkconfig.CMEKProtocolTag:            strings.ToLower(cmekConfig.Cloud),
			quarkconfig.CMEKProjectIDTag:           cmekConfig.ProjectID,
			quarkconfig.CMEKKeyRingTag:             cmekConfig.KeyRing,
			quarkconfig.CMEKKeyRingLocationTag:     cmekConfig.KeyRingLocation,
			quarkconfig.CMEKKeyNameTag:             cmekConfig.KeyName,
			quarkconfig.CMEKTimeoutTag:             quarkconfig.CMEKDefaultTimeout,
			quarkconfig.CMEKRootCACertMountPathTag: quarkconfig.RootCACertMountPath,
		},
	}

	// If CMEK Emulator config is set then set the Oauth and verify values as well
	if cmekEmulatorConfig != nil && strings.TrimSpace(cmekEmulatorConfig.OauthURL) != "" {
		cmekConfigMap.Data[quarkconfig.CMEKEmulatorOauthTag] = cmekEmulatorConfig.OauthURL
	}
	if cmekEmulatorConfig != nil && strings.TrimSpace(cmekEmulatorConfig.VerifyIpHost) != "" {
		cmekConfigMap.Data[quarkconfig.CMEKEmulatorVerifyIpHostTag] = cmekEmulatorConfig.VerifyIpHost
	}

	// If a valid log level is specified then set it in the CMEK configmap
	if len(strings.TrimSpace(cmekConfig.LogLevel)) != 0 {
		if isValidCMEKLogLevel(cmekConfig.LogLevel) {
			cmekConfigMap.Data[quarkconfig.CMEKLogLevelTag] = strings.ToLower(cmekConfig.LogLevel)
		} else {
			errMsg = "CMEK log level should be one of debug|info|notice|warning|error|critical|alert|emergency"
			log.WithFields(log.Fields{
				"storagePool": storagePoolName,
			}).Error(errMsg)
			return nil, fmt.Errorf(errMsg)
		}
	}

	cmekConfigMapCreated, err := client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Create(context.TODO(), cmekConfigMap, metav1.CreateOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePoolName,
			"configMap":   cmekConfigMap.Name,
		}).Error("Failed to create cmek config map: ", err)
		return nil, err
	}
	return cmekConfigMapCreated, nil
}

func (client *NVCClient) NetAppStoragePoolUpdateSubnets(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	ilbSubnets []string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if len(ilbSubnets) == 0 {
		return nil, fmt.Errorf("no subnets specified, please specify one or more subnets")
	}
	storagePool.Spec.IlbSubnets = ilbSubnets
	updatedStoragePool, err := client.netAppStoragePoolUpdate(storagePool)
	return updatedStoragePool, err
}

func (client *NVCClient) NetAppStoragePoolEnable(storagePool *quarkv1alpha1.NetAppStoragePool) (*quarkv1alpha1.NetAppStoragePool, error) {
	if storagePool.ObjectMeta.DeletionTimestamp != nil {
		return storagePool, fmt.Errorf("failed to enable storage pool, getting deleted")
	}
	storagePool.Spec.Disabled = false
	return client.netAppStoragePoolUpdate(storagePool)
}

func (client *NVCClient) NetAppStoragePoolDisable(storagePool *quarkv1alpha1.NetAppStoragePool) (*quarkv1alpha1.NetAppStoragePool, error) {
	if storagePool.ObjectMeta.DeletionTimestamp != nil {
		return storagePool, fmt.Errorf("failed to disable storage pool, getting deleted")
	}
	storagePool.Spec.Disabled = true
	return client.netAppStoragePoolUpdate(storagePool)
}

func (client *NVCClient) NetAppStoragePoolKerberosConfigCreate(
	sp *quarkv1alpha1.NetAppStoragePool,
	adUsername, adPassword string,
	ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration,
	KerbServerConfiguration quarkv1alpha1.KerbServerConfiguration,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(sp.Name)
	if storagePool == nil || err != nil {
		return nil, err
	}

	if kerbEnabled, err := client.IsNFS41FeatureEnabled(); !kerbEnabled {
		return nil, fmt.Errorf("Kerberos feature is not enabled: %w", err)
	}

	// Blocking if KerbServerConfiguration pointer is not yet initialized in CRD.
	// Initialization is done in podrick controller, during storagepool sync after upgrade.
	// This condition is added to allow kerberos config create only with successfully upgraded SP.
	if storagePool.Spec.KerbServerConfiguration == nil {
		return nil, fmt.Errorf("Storagepool is not online or not upgraded to kerberos supported version.")
	}

	// Check if kerberos already configured.
	if util.KerberosConfigurationInitialized(storagePool) {
		return nil, fmt.Errorf("Kerberos is already configured.")
	}

	if err = util.ValidateKerberosConfiguration(KerbServerConfiguration); err != nil {
		return nil, err
	}

	if client.NetAppStoragePoolSMBInitialized(storagePool) == false {
		//Creating SMB config. SMB configuration is required for Kerberos.
		storagePool, err = client.NetAppStoragePoolSMBConfigCreate(storagePool, adUsername, adPassword, ADDNSIPs, SMBServerConfiguration, SMBADConfiguration)
		if err != nil {
			return nil, fmt.Errorf("Failed to create SMB config: %w", err)
		}
	} else {
		if storagePool.Status.SMBADConfiguration != SMBADConfiguration ||
			!util.IsEqualStrings(storagePool.Status.ADDNSIPs, ADDNSIPs) {
			return nil, fmt.Errorf("Updating DNS and AD config is not allowed.")
		}

		//Modify SMB Config.
		if storagePool.Status.SMBServerConfiguration != SMBServerConfiguration {
			storagePool, err = client.NetAppStoragePoolSMBServerConfigurationUpdate(storagePool, adUsername, adPassword, SMBServerConfiguration)
			if err != nil {
				return nil, fmt.Errorf("Failed to update SMB config: %w", err)
			}
		}

	}

	secretName := util.GetVolCreateADSecretName(storagePool.Name + "-kerb-")
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, adUsername, adPassword)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create AD secret.")
		return nil, err
	}

	kerbServerConfiguration := quarkv1alpha1.KerbServerConfiguration{
		KDCServerIP:  KerbServerConfiguration.KDCServerIP,
		ADServerName: KerbServerConfiguration.ADServerName,
	}
	storagePool.Spec.KerbServerConfiguration = &kerbServerConfiguration

	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Set the forceRetry field to true in the CRD error condition to prevent the subsequent wait_for
	// API from treating outdated error as a fatal error. Condition clearing happens from controller(ccpd) after
	// operation is completed.
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolKerbConfigurationError)
	}
	return storagePool, err
}

func (client *NVCClient) NetAppStoragePoolKerberosConfigCreateAbort(
	sp *quarkv1alpha1.NetAppStoragePool,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(sp.Name)
	if storagePool == nil || err != nil {
		return nil, err
	}

	if kerbEnabled, err := client.IsNFS41FeatureEnabled(); !kerbEnabled {
		return nil, fmt.Errorf("Kerberos feature is not enabled: %w", err)
	}

	//Fail abort if Kerberos is already configured.
	if util.KerberosConfigurationInitialized(storagePool) {
		return nil, fmt.Errorf("Failed to abort kerberos config: Kerberos is already configured.")
	}

	// Abort SMB config create if SMB is not ready yet.
	if client.NetAppStoragePoolSMBInitialized(storagePool) == false {
		storagePool, err = client.NetAppStoragePoolSMBConfigCreateAbort(storagePool)
		if err != nil {
			return nil, fmt.Errorf("Failed to abort kerberos config creation: %w", err)
		}
	} else {
		//Log the details and continue with the abort operation as we need to handle the Kerberos configuration anyways.
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"netbios":     storagePool.Status.SMBServerConfiguration.Netbios,
			"domain":      storagePool.Status.SMBServerConfiguration.Domain,
		}).Info("Kerberos configuration creation couldn't be fully aborted. SMB configuration won't be removed as its already configured.")
	}

	storagePool.Spec.KerbServerConfiguration = &quarkv1alpha1.KerbServerConfiguration{}
	storagePool.Spec.KerbConfigUpdateCount += 1
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleaning up the existing fatal error from CRD error conditions to ensure the subsequent wait_for_aborted API
	// does not return outdated fatal errors. Since the Kerberos abort operation is always guaranteed to be completed by
	// the controller (ccpd), it is safe to remove the fatal error condition entirely at this point.
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolKerbConfigurationError)
	}

	return storagePool, err
}

func (client *NVCClient) NetAppStoragePoolADConfigUpdate(
	sp *quarkv1alpha1.NetAppStoragePool,
	adUsername, adPassword string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	KerbServerConfiguration quarkv1alpha1.KerbServerConfiguration,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	updateKerbSec := false
	storagePool, err := client.NetAppStoragePoolGet(sp.Name)
	if storagePool == nil || err != nil {
		return nil, err
	}

	if kerbEnabled, err := client.IsNFS41FeatureEnabled(); !kerbEnabled {
		return nil, fmt.Errorf("Kerberos feature is not enabled: %w", err)
	}

	if err = util.ValidateKerberosConfiguration(KerbServerConfiguration); err != nil {
		return nil, err
	}

	if storagePool.Spec.KerbServerConfiguration == nil || storagePool.Status.KerbServerConfigurationStatus == nil ||
		storagePool.Status.KerbServerConfigurationStatus.KDCServerIP == "" ||
		storagePool.Status.KerbServerConfigurationStatus.ADServerName == "" {
		return nil, fmt.Errorf("Updating Kerberos configuration is not allowed: Kerberos not yet configured")
	}

	secretName := util.GetVolModifyADSecretName(storagePool.Name + "-kerb-")
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, adUsername, adPassword)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to update AD secret.")
		return nil, err
	}

	//Update Kerberos config
	if KerbServerConfiguration.ADServerName != storagePool.Spec.KerbServerConfiguration.ADServerName ||
		KerbServerConfiguration.KDCServerIP != storagePool.Spec.KerbServerConfiguration.KDCServerIP {
		storagePool.Spec.KerbServerConfiguration = &KerbServerConfiguration
		updateKerbSec = true
	}

	// Update SMB config
	// NetAppStoragePoolSMBServerConfigurationUpdate internally calls netAppStoragePoolUpdate
	if SMBServerConfiguration != storagePool.Status.SMBServerConfiguration {
		storagePool.Spec.KerbConfigUpdateCount += 1
		storagePool, err = client.NetAppStoragePoolSMBServerConfigurationUpdate(storagePool, adUsername, adPassword, SMBServerConfiguration)
	} else if updateKerbSec {
		storagePool.Spec.KerbConfigUpdateCount += 1
		storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	}

	if err != nil {
		return nil, fmt.Errorf("Failed to update SMB/Kerberos config: %w", err)
	}

	// Set the forceRetry field to true in the CRD error condition to prevent the subsequent wait_for
	// API from treating outdated error as a fatal error. Condition clearing happens from controller(ccpd) after
	// operation is completed.
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolKerbConfigurationError)
	}

	return storagePool, err
}

// NetAppStoragePoolKerberosReady Returns "true" if Kerberos is configured and ready for use in storagepool.
// otherwise returns "false" along with error.
func (client *NVCClient) NetAppStoragePoolKerberosReady(sp *quarkv1alpha1.NetAppStoragePool) (bool, error) {
	if kerbEnabled, err := client.IsNFS41FeatureEnabled(); !kerbEnabled {
		return false, err
	}

	if sp.Spec.KerbServerConfiguration == nil ||
		sp.Spec.KerbServerConfiguration.ADServerName == "" ||
		sp.Spec.KerbServerConfiguration.KDCServerIP == "" {
		return false, fmt.Errorf("Kerberos configuration not found for storagepool: %s", sp.Name)
	}

	if !client.NetAppStoragePoolSMBReady(sp) {
		return false, fmt.Errorf("SMB configuration not ready for storagepool: %s", sp.Name)
	}

	if errPresent, err := util.KerberosConfigurationErrorPresent(sp); errPresent {
		return false, fmt.Errorf("Kerberos configuration not ready for storagepool: %s, err: %s", sp.Name, err)
	}

	isEqual, err := util.KerberosConfigurationCompareEqual(sp, false)
	if !isEqual {
		return false, fmt.Errorf("Kerberos configuration not ready for storagepool: %s, err: %s", sp.Name, err)
	}

	return true, nil
}

func (client *NVCClient) NetAppStoragePoolKerberosAborted(sp *quarkv1alpha1.NetAppStoragePool) (bool, error) {
	if kerbEnabled, err := client.IsNFS41FeatureEnabled(); !kerbEnabled {
		return false, err
	}

	if errPresent, errStr := util.KerberosConfigurationErrorPresent(sp); errPresent {
		return false, fmt.Errorf("Kerberos config abort in-progress for storagepool: %s, err: %s", sp.Name, errStr)
	}

	isEqual, err := util.KerberosConfigurationCompareEqual(sp, true)
	if !isEqual {
		return false, fmt.Errorf("Kerberos config abort in-progress for storagepool: %s, err: %s", sp.Name, err)
	}

	return true, nil
}

func (client *NVCClient) IsNFS41FeatureEnabled() (bool, error) {

	if !client.EnableCCPd {
		return false, fmt.Errorf("CCPd is not enabled in quark")
	}
	if !client.EnableNfs41 {
		return false, fmt.Errorf("NFS4.1 is not enabled in quark")
	}
	// TODO: Check EnableGcp1p once we enable 1P only for NFS4.1

	return true, nil
}

// isNetAppStoragePoolEncryptionEnabled checks if cmek-<sp-uuid> label is available
func isNetAppStoragePoolEncryptionEnabled(storagePool *quarkv1alpha1.NetAppStoragePool) (ok bool) {
	_, ok = storagePool.Labels[quarkconfig.CMEKConfigMapLabel]
	return ok
}

// isStoragePoolEncryptionPaused checks if storage pool encryption is set to pause
func isStoragePoolEncryptionPaused(storagePool *quarkv1alpha1.NetAppStoragePool) (ok bool) {
	if value, ok := storagePool.Labels[quarkconfig.StoragePoolEncryptionLabel]; ok {
		return value == quarkconfig.StoragePoolEncryptionPause.String()
	}
	return false
}

// isStoragePoolEncryptionResume check if storage pool encryption is set to resume
func isStoragePoolEncryptionResume(storagePool *quarkv1alpha1.NetAppStoragePool) (ok bool) {
	if value, ok := storagePool.Labels[quarkconfig.StoragePoolEncryptionLabel]; ok {
		return value == quarkconfig.StoragePoolEncryptionResume.String()
	}
	return false
}

func (client *NVCClient) preCheckNetAppStoragePoolEncryption(
	spUUID string,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	// check if the spUUID is an empty string
	if spUUID == "" {
		return nil, fmt.Errorf("storage pool uuid input cannot be empty")
	}
	// get latest storage pool instance
	storagePool, err = client.NetAppStoragePoolGet(spUUID)
	if err != nil {
		return nil, err
	}
	// Check StoragePool Readiness to perform operation
	err = checkNetAppStoragePoolReadiness(client, storagePool)
	if err != nil {
		return nil, fmt.Errorf("NetAppStoragePool Readiness check failed: %s", err.Error())
	}
	// if storage pool encryption is not enabled, return error
	if !isNetAppStoragePoolEncryptionEnabled(storagePool) {
		return nil, fmt.Errorf("encryption is not enabled for storage pool")
	}
	return storagePool, nil
}

func updateNetAppStoragePoolEncryptionValue(client *NVCClient, storagePool *quarkv1alpha1.NetAppStoragePool, value string) (err error) {
	storagePool.Labels[quarkconfig.StoragePoolEncryptionLabel] = value
	_, err = client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		return fmt.Errorf("Failed to %s NetAppStoragePool encryption: %s", value, err)
	}
	return nil
}

func updateNetAppStoragePoolEncryptionResume(client *NVCClient, storagePool *quarkv1alpha1.NetAppStoragePool) (err error) {
	// set storage pool CR with sp encryption label to resume
	return updateNetAppStoragePoolEncryptionValue(client, storagePool, quarkconfig.StoragePoolEncryptionResume.String())
}

func updateNetAppStoragePoolEncryptionPause(client *NVCClient, storagePool *quarkv1alpha1.NetAppStoragePool) (err error) {
	// set storage pool CR with sp encryption label to pause
	return updateNetAppStoragePoolEncryptionValue(client, storagePool, quarkconfig.StoragePoolEncryptionPause.String())
}

func checkNetAppStoragePoolReadiness(client *NVCClient, storagePool *quarkv1alpha1.NetAppStoragePool) (err error) {
	// Check if SP has been marked for Deletion
	if storagePool.ObjectMeta.DeletionTimestamp != nil {
		return fmt.Errorf("storage pool is getting deleted.")
	}
	// Check if storage pool is disabled
	if isDisabled := client.NetAppStoragePoolDisabled(storagePool); isDisabled {
		return fmt.Errorf("storage pool is disabled.")
	}
	// Check if storage pool is not online
	if !client.NetAppStoragePoolOnline(storagePool) {
		return fmt.Errorf("storage pool is not online.")
	}
	return nil
}

// NetAppStoragePoolEncryptionStart Encrypts existing storage pool volumes
func (client *NVCClient) NetAppStoragePoolEncryptionStart(
	storagePoolUUID string,
	cmekConfig *quarkv1alpha1.CMEKConfiguration,
	optionFns ...quarkv1alpha1.StoragePoolOptionsFn,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	// check if the spUUID is an empty string
	if storagePoolUUID == "" {
		return nil, fmt.Errorf("storage pool uuid input cannot be empty")
	}
	// Check cmekConfig input
	if cmekConfig == nil {
		return nil, fmt.Errorf("CMEK Configuration input cannot be nil")
	}
	// Get latest storage pool instance
	storagePool, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
	}
	// Check StoragePool Readiness to perform operation
	err = checkNetAppStoragePoolReadiness(client, storagePool)
	if err != nil {
		return nil, fmt.Errorf("NetAppStoragePool Readiness check failed: %s", err.Error())
	}
	// Check if the storage pool already has the cmek label, if so, throw error,
	if existingCMEKConfigName, ok := storagePool.Labels[quarkconfig.CMEKConfigMapLabel]; ok && len(existingCMEKConfigName) > 0 {
		return nil, fmt.Errorf("storage pool already has cmek label %s=%s", quarkconfig.CMEKConfigMapLabel, existingCMEKConfigName)
	}
	// Check for additional SP config - Emulator
	additionalConfig := quarkv1alpha1.AdditionalStoragePoolConfig{}
	for _, optionFn := range optionFns {
		optionFn(&additionalConfig)
	}
	// Create CMEK ConfigMap
	var cmekConfigMap *corev1.ConfigMap
	if cmekConfigMap, err = client.createCmekConfigMap(storagePool.Name, cmekConfig, additionalConfig.CMEKEmulatorConfig); err != nil {
		return nil, err
	}
	storagePool.Labels[quarkconfig.CMEKConfigMapLabel] = cmekConfigMap.Name

	// Update Storage Pool with CMEK config reference
	err = updateCMEKConfigMapOwnerRef(client, storagePool, cmekConfigMap)
	if err != nil {
		return nil, err
	}

	return client.netAppStoragePoolUpdate(storagePool)
}

// NetAppStoragePoolEncryptionPause pauses storage pool encryption
func (client *NVCClient) NetAppStoragePoolEncryptionPause(
	storagePoolUUID string,
) (err error) {
	/*
		(1) if sp encryption is not enabled - return error - cannot pause encryption for unencrypted SP
		(2) if CMEK Config is invalid or encryption is not yet initialized - return err
		(3) if CMEK does not have encryption state set or if the encryption state is set to completed, return error
		(4) if sp encryption label is set to pause - no-op - return
		(5) if sp encryption label is set to resume or not set - update the label value to pause and return
	*/
	// condition (1) if sp not encrypted, return error
	storagePool, err := client.preCheckNetAppStoragePoolEncryption(storagePoolUUID)
	if err != nil {
		return fmt.Errorf("%s: %s", encryptPauseErr, err)
	}

	// condition (2) if CMEK Config is invalid or encryption is not yet initialized, return error
	if err = client.isCMEKErrCondPresentOrEncryptionNotInitialized(storagePool, quarkconfig.StoragePoolEncryptionPause.String()); err != nil {
		return err
	}

	// condition (3) if CMEK does not have encryption state set or if the encryption state is set to completed, return error
	if err = client.checkValidEncryptionStateForPauseResume(storagePool, quarkconfig.StoragePoolEncryptionPause.String()); err != nil {
		return err
	}

	// condition (4) check if the storage pool is already set to pause, then no action required
	if isStoragePoolEncryptionPaused(storagePool) {
		log.Debug("storage pool encryption is already set to pause")
		return nil
	}
	// condition (5) if label is not set or it might have been set to resume, so update the value
	// to pause and return
	return updateNetAppStoragePoolEncryptionPause(client, storagePool)
}

// NetAppStoragePoolEncryptionResume resumes storage pool encryption which is in paused state
func (client *NVCClient) NetAppStoragePoolEncryptionResume(
	storagePoolUUID string,
) (err error) {
	/*
		(1) if sp encryption is not enabled - return error - cannot resume encryption for unencrypted SP
		(2) if CMEK Config is invalid or encryption is not yet initialized - return err
		(3) if CMEK does not have encryption state set or if the encryption state is set to completed, return err
		(4) if sp encryption label is set to resume already - no-op - return nil
		(5) if sp encryption label is set to pause - updates the label value to resume and return
	*/

	// condition (1) if sp not encrypted, return error
	storagePool, err := client.preCheckNetAppStoragePoolEncryption(storagePoolUUID)
	if err != nil {
		return fmt.Errorf("%s: %s", encryptResumeErr, err)
	}

	// condition (2) if CMEK Config is invalid or encryption is not yet initialized, return error
	if err = client.isCMEKErrCondPresentOrEncryptionNotInitialized(storagePool, quarkconfig.StoragePoolEncryptionResume.String()); err != nil {
		return err
	}

	// condition (3) if CMEK does not have encryption state set or if the encryption state is set to completed, return error
	if err = client.checkValidEncryptionStateForPauseResume(storagePool, quarkconfig.StoragePoolEncryptionResume.String()); err != nil {
		return err
	}

	// condition (4) check if the storage pool is set to resume already, no action required
	if isStoragePoolEncryptionResume(storagePool) {
		log.Debug("storage pool encryption is already set to resume")
		return nil
	}

	// condition (5) if storage pool encryption is set to pause, update to resume and return
	if isStoragePoolEncryptionPaused(storagePool) {
		log.Debug("storage pool encryption is currently set to pause, setting to resume")
		return updateNetAppStoragePoolEncryptionResume(client, storagePool)
	}

	// return error "Cannot resume a storage pool encryption not in a paused state"
	return fmt.Errorf("%s: storage pool is currently not in pause", encryptResumeErr)
}

func (client *NVCClient) deleteCMEKConfigMap(cmekConfigMap string) error {
	err := client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Delete(context.TODO(), cmekConfigMap, metav1.DeleteOptions{})
	if err != nil && !errors.IsNotFound(err) {
		return err
	}

	return nil
}

// NetAppStoragePoolEncryptionUpdate updates the storage pool encryption configuration
func (client *NVCClient) NetAppStoragePoolEncryptionUpdate(
	storagePoolUUID string,
	cmekConfig *quarkv1alpha1.CMEKConfiguration,
) (err error) {
	// Check Storage Pool UUID
	if storagePoolUUID == "" {
		return fmt.Errorf("storage pool UUID not specified")
	}

	// Check cmekConfig input
	if cmekConfig == nil {
		return fmt.Errorf("CMEK Configuration not specified")
	}

	// Get storage pool object
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return fmt.Errorf("failed to get the storage pool: %v", err.Error())
	}

	// Check StoragePool Readiness to perform operation
	err = checkNetAppStoragePoolReadiness(client, storagePool)
	if err != nil {
		return fmt.Errorf("NetAppStoragePool Readiness check failed: %s", err.Error())
	}

	// Check if the storage pool has CMEK configured? i.e. CMEK ConfigMap label is present
	// or the quark pod was deployed with an CMEK ConfigMap earlier and something
	// failed resulting in label deletion
	label, ok := storagePool.Labels[quarkconfig.CMEKConfigMapLabel]
	if !ok {
		return fmt.Errorf("storage pool encryption is not enabled")
	}

	// This is the reconcile logic for NVC only when this completes, the update API should be invoked
	if storagePool.Status.LastEncryptionConfiguration != label {
		return fmt.Errorf("storage pool encryption initialization did not complete yet, encryption update can be initiated only if its completed")
	}

	// Check if there is any ongoing storage pool encryption
	// Even if encryption is set to scheduled which means that podrick is yet to
	// take action, the volume objects have also been labeled to scheduled which
	// will bound to get encrypted. We should not allow change of CMEK Configuration
	// in this state as well
	acceptableEncryptionStates := []string{
		quarkv1alpha1.EncryptionStateCompleted.String(),
		quarkv1alpha1.EncryptionStateScheduled.String(),
	}
	if len(storagePool.Status.Encryption.String()) > 0 &&
		!util.Contains(storagePool.Status.Encryption.String(), acceptableEncryptionStates) {
		return fmt.Errorf("storage pool encryption is ongoing, please try later")
	}

	// Check if CMEKConfigurationError condition is present
	configErrCondPresent, _ := util.IsErrorConditionPresent(
		storagePool.Status.Conditions,
		quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)
	if configErrCondPresent {
		// If SVM Key was already created, but CMEKConfigError condition is present
		// indicates that key manager encountered an error on boot up
		if storagePool.Status.EncryptionInitialized {
			errMsg := (util.GetNetAppStoragePoolCondition(storagePool,
				quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)).Message
			return fmt.Errorf(
				"storage pool has CMEK configuration error, please reach out to support: %s",
				errMsg)
		}
	} else {
		// No CMEK Config error condition present, allow change only if SVM key
		// was not initialized
		// Check if SVM key was created, in which case CMEK configuration change
		// is considered as a rekey operation and not supported for 24.01
		if storagePool.Status.EncryptionInitialized {
			return fmt.Errorf(
				`storage pool encryption already initialized,
				rekey operation is not supported for this release`)
		}
	}

	// No error or unsupported case encountered, update Storage Pool with new CMEK ConfigMap
	// Create CMEK ConfigMap
	var cmekConfigMap *corev1.ConfigMap
	if cmekConfigMap, err = client.createCmekConfigMap(storagePool.Name, cmekConfig, nil); err != nil {
		return err
	}

	// Save old CMEK ConfigMap name
	// Here we are giving precedence to CMEK ConfigMap from label to account
	// for the case where NVC didn't act on the CMEK ConfigMap label to update
	// quark pod deployment
	oldCMEKConfigMapName := storagePool.Status.LastEncryptionConfiguration
	if occn, ok := storagePool.Labels[quarkconfig.CMEKConfigMapLabel]; ok {
		oldCMEKConfigMapName = occn
	}

	// Update Strorage Pool CMEK label to reflect new CMEK ConfigMap
	storagePool.Labels[quarkconfig.CMEKConfigMapLabel] = cmekConfigMap.Name

	_, err = client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		if _, ok := err.(StaleVersionError); ok {
			return err
		}
		return fmt.Errorf(
			"failed to add CMEK ConfigMap label %s=%s on storage pool: %v",
			quarkconfig.CMEKConfigMapLabel, cmekConfigMap.Name, err)
	}

	// Add owner reference on new CMEK ConfigMap
	err = updateCMEKConfigMapOwnerRef(client, storagePool, cmekConfigMap)
	if err != nil {
		return err
	}

	// Delete the old CMEK ConfigMap
	if len(oldCMEKConfigMapName) > 0 {
		// Delete CMEK Configmap, if ConfigMap is already deleted, then its a no-op
		err := client.deleteCMEKConfigMap(oldCMEKConfigMapName)
		if err != nil {
			return fmt.Errorf(
				"failed to delete CMEK ConfigMap for provided storage pool: %v",
				corev1.ErrUnexpectedEndOfGroupGenerated)
		}
	} else {
		return fmt.Errorf(
			"failed to find the CMEK ConfigMap label %s on provided storage pool",
			quarkconfig.CMEKConfigMapLabel)
	}

	return nil
}

// NetAppStoragePoolVolumesEncryptionStatus to get volumes of a Storage pool in a specific encryption state
func (client *NVCClient) NetAppStoragePoolVolumesEncryptionStatus(
	storagePoolUUID string, optionFns ...quarkv1alpha1.VolumeFiltersOptionsFn,
) ([]*quarkv1alpha1.NetAppVolumeEncryptionStatus, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetAppStoragePool: %v", err)
	}

	// Check StoragePool Readiness to perform operation
	err = checkNetAppStoragePoolReadiness(client, storagePool)
	if err != nil {
		return nil, fmt.Errorf("NetAppStoragePool Readiness check failed: %s", err.Error())
	}

	// Check if CMEKConfigurationError condition is present
	configErrCondPresent, _ := util.IsErrorConditionPresent(
		storagePool.Status.Conditions,
		quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)
	if configErrCondPresent {
		errMsg := (util.GetNetAppStoragePoolCondition(storagePool,
			quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)).Message

		// Check whether is an new CMEK initialization issue or post reboot
		if storagePool.Status.EncryptionInitialized {
			err = fmt.Errorf(
				"failed to reinitialize encryption: %v",
				errMsg)
		} else {
			err = fmt.Errorf(
				"failed to initialize encryption due to invalid configuration: %v",
				errMsg)
		}
		return nil, err
	}

	acceptableEncryptionStates := []string{
		quarkv1alpha1.EncryptionStateCompleted.String(),
		quarkv1alpha1.EncryptionStateScheduled.String(),
		quarkv1alpha1.EncryptionStatePaused.String(),
		quarkv1alpha1.EncryptionStateInProgress.String(),
		quarkv1alpha1.EncryptionStateFailed.String(),
	}

	/* This validation is required as this API is for migration purpose */
	// Check if the storage pool has the cmek label, if not, throw error
	if _, ok := storagePool.Labels[quarkconfig.CMEKConfigMapLabel]; ok {
		if len(storagePool.Status.Encryption) == 0 {
			return nil, fmt.Errorf("encryption status is only applicable for plain-text volume conversion and this storage pool (%s) is/was created with encryption", storagePoolUUID)
		}
	} else {
		return nil, fmt.Errorf("failed to get encryption status as encryption is not enabled in NetAppStoragePool: %s", storagePoolUUID)
	}

	// Check for additional SP config - Encryption State
	additionalConfig := quarkv1alpha1.AdditionalVolumeFilters{}
	for _, optionFn := range optionFns {
		optionFn(&additionalConfig)
	}
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.StoragePoolLabel] = storagePoolUUID

	if additionalConfig.VolumeUUID != "" && additionalConfig.EncryptionState.String() != "" {
		return nil, fmt.Errorf("either provide volume UUID or the encryption state but not both, its not supported")
	}

	if additionalConfig.VolumeUUID != "" {
		volume, err := client.NetAppVolumeGet(additionalConfig.VolumeUUID)
		if err != nil {
			return nil, fmt.Errorf("failed to get the volume %s for NetAppStoragePool: %v", additionalConfig.VolumeUUID, err)
		}
		// Check if Vol is part of the given SP
		if val, ok := volume.Labels[quarkconfig.StoragePoolLabel]; ok {
			if val != storagePoolUUID {
				return nil, fmt.Errorf("Volume is not part of given NetAppStoragePool")
			}
		}
		selectorSet[quarkconfig.VolumeLabel] = additionalConfig.VolumeUUID
	}

	if additionalConfig.EncryptionState.String() != "" {
		if !util.Contains(additionalConfig.EncryptionState.String(), acceptableEncryptionStates) {
			return nil, fmt.Errorf("%s is not the valid state ", additionalConfig.EncryptionState.String())
		}
		selectorSet[quarkconfig.VolumeEncryptionLabel] = additionalConfig.EncryptionState.String()
	}

	listOptions := metav1.ListOptions{
		LabelSelector: selectorSet.String(),
	}

	result, err := client.listWithRetry(&quarkv1alpha1.NetAppVolumeList{}, listOptions)
	if err != nil {
		return nil, fmt.Errorf("failed to get the volumes for NetAppStoragePool: %v", err)
	}
	volumes := result.(*quarkv1alpha1.NetAppVolumeList)
	volsWithState := make([]*quarkv1alpha1.NetAppVolumeEncryptionStatus, len(volumes.Items))

	for i, vol := range volumes.Items {
		volsWithState[i] = &quarkv1alpha1.NetAppVolumeEncryptionStatus{
			VolumeUUID:       vol.Status.VolumeUUID,
			EncryptionStatus: vol.Status.EncryptionStatus,
		}
	}

	return volsWithState, nil
}

// NetAppStoragePoolEncryptionStatus, get the encryption status of a Storage pool
func (client *NVCClient) NetAppStoragePoolEncryptionStatus(
	storagePoolUUID string,
) (quarkv1alpha1.EncryptionState, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return "", fmt.Errorf("failed to get the NetAppStoragePool: %v", err)
	}

	// Check StoragePool Readiness to perform operation
	err = checkNetAppStoragePoolReadiness(client, storagePool)
	if err != nil {
		return "", fmt.Errorf("NetAppStoragePool Readiness check failed: %s", err.Error())
	}

	// Check if CMEKConfigurationError condition is present
	configErrCondPresent, _ := util.IsErrorConditionPresent(
		storagePool.Status.Conditions,
		quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)
	if configErrCondPresent {
		// In migration case, its always beneficial to return the current encryption state when there is config error
		// If not immediately seen, on restart of dmap if cmekConfigError condition is added to Storage pool's status
		// then, this Status API could return current state along with the error
		// By design, on first boot and configError, state is "scheduled"
		errMsg := (util.GetNetAppStoragePoolCondition(storagePool,
			quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)).Message
		// Check whether is an new CMEK initialization issue or post reboot
		if storagePool.Status.EncryptionInitialized {
			err = fmt.Errorf(
				"failed to reinitialize encryption: %v",
				errMsg)
		} else {
			err = fmt.Errorf(
				"failed to initialize encryption due to invalid configuration: %v",
				errMsg)
		}
		return storagePool.Status.Encryption, err
	}

	// Check if the storage pool has the cmek label, if not, throw error
	if _, ok := storagePool.Labels[quarkconfig.CMEKConfigMapLabel]; ok {
		if len(storagePool.Status.Encryption) == 0 {
			return "", fmt.Errorf("encryption status is only applicable for plain-text volume conversion and this storage pool (%s) is/was created with encryption", storagePoolUUID)
		}
	} else {
		return "", fmt.Errorf("failed to get encryption status as encryption is not enabled in NetAppStoragePool: %s", storagePoolUUID)
	}

	return storagePool.Status.Encryption, nil
}

func (client *NVCClient) NetAppStoragePoolEncryptionInitialization(sp *quarkv1alpha1.NetAppStoragePool) bool {
	if label, ok := sp.Labels[quarkconfig.CMEKConfigMapLabel]; ok {
		return sp.Status.EncryptionInitialized && label == sp.Status.LastEncryptionConfiguration
	}

	return false
}

// isCMEKErrCondPresentOrEncryptionNotInitialized will check 2 conditions
// (1) if CMEKConfigError condition is present, return err
// (2) if encryption is not initialized, return err
func (client *NVCClient) isCMEKErrCondPresentOrEncryptionNotInitialized(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	op string,
) error {
	// condition (1), if CMEKConfigError condition is present,
	// pause or resume should be blocked, return err
	configErrCondPresent, _ := util.IsErrorConditionPresent(
		storagePool.Status.Conditions,
		quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)
	if configErrCondPresent {
		return fmt.Errorf("cannot %s, encryption configuration provided for storage pool is not valid.", op)
	} else {
		// condition (2), if encryption is not initialized,
		// pause or resume should be blocked, return err
		if !storagePool.Status.EncryptionInitialized {
			return fmt.Errorf("cannot %s, storage pool encryption is not initialized yet.", op)
		}
	}
	return nil
}

// checkValidEncryptionStateForPauseResume will check 2 conditions
// (1) if encryption field is not set in Storage Pool, return err
// (2) if encryption is set to completed, return err
func (client *NVCClient) checkValidEncryptionStateForPauseResume(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	op string,
) error {
	unacceptableState := []string{
		quarkv1alpha1.EncryptionStateCompleted.String(),
	}
	// condition  (1), if encryption field is not set in Storage Pool return error
	// as this represents the Storage Pool in which CMEK was set at creation time.
	if len(storagePool.Status.Encryption.String()) == 0 {
		return fmt.Errorf("%s is only applicable for plain-text volume conversion and this storage pool (%s) is/was created with encryption", op, storagePool.Name)
	} else {
		// Condition (2), if encryption field is set to completed, encryption cannot
		// be paused or resumed, return error
		if util.Contains(storagePool.Status.Encryption.String(), unacceptableState) {
			return fmt.Errorf("cannot %s, as encryption is completed for storage pool: %s", op, storagePool.Name)
		}
	}
	return nil
}
