package util

import (
	"fmt"
	"strings"
	"time"

	"quark.netapp.io/volume-controller/pkg/quarkconfig"

	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	// Quark
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

const SubsystemInternalErrorPrifix = "Internal: "

func SetStoragepoolSubsystemEvent(module string, status corev1.ConditionStatus) *corev1.Event {
	if module == quarkconfig.BootSubsystem {
		return nil
	}
	eventType, eventState := corev1.EventTypeNormal, quarkconfig.SubsystemReady
	if status == corev1.ConditionFalse {
		eventType, eventState = corev1.EventTypeWarning, quarkconfig.SubsystemNotReady
	}
	eventReason, eventMsg := quarkconfig.NetAppStoragePoolSubsystemStatusChanged, fmt.Sprintf("StoragePool %s subsystem state changed to %s", module, eventState)

	return &corev1.Event{
		Type:    eventType,
		Reason:  eventReason,
		Message: eventMsg,
	}
}

// returns true if storage pool online condition exists
func IsStoragePoolHealthInitialized(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return getCondition(storagePool.Status.Conditions, quarkv1alpha1.NetAppStoragePoolOnline) != nil
}

// returns true if granular health is enabled and is initialized in storage pool
func UseStoragePoolGranularHealth(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return storagePool.Status.GranularHealth != nil
}

// returns storagepool online from online condition(if present)
func IsNetAppStoragePoolOnline(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	onlineCondition := getCondition(storagePool.Status.Conditions, quarkv1alpha1.NetAppStoragePoolOnline)

	return onlineCondition != nil && onlineCondition.Status == corev1.ConditionTrue
}

// Mark storagePool as offline if it is already online. When pod restarts, podrick will mark storagePool as online
func MakeNetAppStoragePoolOffline(storagePool *quarkv1alpha1.NetAppStoragePool, msg string) *quarkv1alpha1.NetAppStoragePool {
	conditions := storagePool.Status.Conditions

	currTime := metav1.NewTime(time.Now())
	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.NetAppStoragePoolOnline && conditions[i].Status == corev1.ConditionTrue {
			conditions[i].LastHeartbeatTime = currTime
			conditions[i].LastTransitionTime = currTime
			conditions[i].Status = corev1.ConditionFalse
			conditions[i].Reason = quarkconfig.NetAppStoragePoolNotReady
			conditions[i].Message = msg
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"conditions":  conditions,
			}).Debug("Updated online condition.")
			break
		}
	}
	storagePool.Status.Conditions = conditions

	// Only if Granular health is enabled
	if storagePool.Status.GranularHealth != nil {
		granularHealth := make(quarkv1alpha1.StoragePoolGranularHealthMap)
		for subsystem := range storagePool.Status.GranularHealth {
			// All the errors causing SP marking to offline will be internal errors.
			// Hence converting to msg in to internal error.
			if !IsInternalSubsystemError(msg) {
				msg = MakeInternalSubsystemError(msg)
			}

			// MArking each subsystem as notReady.
			granularHealth[subsystem] = quarkv1alpha1.NetAppStoragePoolSubsystem{Status: corev1.ConditionFalse, Message: msg, LastTransitionTime: currTime}
		}
		storagePool.Status.GranularHealth = granularHealth
	}

	return storagePool
}

func isSubsystemHealthy(storagePool *quarkv1alpha1.NetAppStoragePool, subsystem string) bool {
	if subsystemHealth, ok := storagePool.Status.GranularHealth[subsystem]; ok {
		return subsystemHealth.Status == corev1.ConditionTrue
	}
	return false
}

// GetSubsystemName returns storagepool subsystem based on volume protocol.
func GetSubsystemName(storagePool *quarkv1alpha1.NetAppStoragePool, volumeProtocol string) string {
	if CheckNfsVolume(volumeProtocol) {
		if storagePool.Status.LdapEnabled {
			return quarkconfig.NfsLdapSubsystem
		} else {
			// Kerberos status is not handled in this release
			// Default NFS status is same as BOOT subsystem status
			return quarkconfig.BootSubsystem
		}
	}
	return quarkconfig.SmbSubsystem
}

// GetVolumeSubsystemStatus returns storage pool subsystem details based on volume protocol.
func GetVolumeSubsystemStatus(storagePool *quarkv1alpha1.NetAppStoragePool, volumeProtocol string) (subsystem string, subsystemReady corev1.ConditionStatus, subsystemMessage string) {
	subsystem, subsystemReady = GetSubsystemName(storagePool, volumeProtocol), corev1.ConditionTrue

	// If the Storagepool is Offline then all the subsystems will be in notReady state. But in this case Podrick/DMAP can be down.
	// Hence these subsystems status in Storagepool CR may not have updated to notReady and the current state could be a stale one.
	// So, setting the subsystem status as offline if Storagepool is offline
	if onlineCondition := getCondition(storagePool.Status.Conditions, quarkv1alpha1.NetAppStoragePoolOnline); onlineCondition != nil && onlineCondition.Status != corev1.ConditionTrue {
		subsystemReady, subsystemMessage = corev1.ConditionFalse, onlineCondition.Message
	} else {
		// get subsystem details from granular health data.
		if subsystemHealth, ok := storagePool.Status.GranularHealth[subsystem]; ok {
			subsystemReady, subsystemMessage = subsystemHealth.Status, subsystemHealth.Message
		}
	}

	return subsystem, subsystemReady, subsystemMessage
}

func IsInternalSubsystemError(message string) bool {
	return strings.HasPrefix(message, SubsystemInternalErrorPrifix)
}

func TrimInternalSubsystemError(message string) string {
	return strings.TrimPrefix(message, SubsystemInternalErrorPrifix)
}

func MakeInternalSubsystemError(message string) string {
	return SubsystemInternalErrorPrifix + message
}
