package util

import (
	"errors"
	"fmt"
	"math"
	"net"
	"path/filepath"
	"reflect"
	"sort"
	"strings"

	"k8s.io/apimachinery/pkg/api/resource"
	"k8s.io/apimachinery/pkg/labels"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"

	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"

	// Quark
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

// StringSetDifference returns the difference between two sets (sets can't have
// duplicates): all the elements in the set a that are not in set b.
func StringSetDifference(a, b []string) (diff []string) {
	m := make(map[string]bool)

	for _, item := range b {
		m[item] = true
	}

	for _, item := range a {
		if !m[item] {
			diff = append(diff, item)
		}
	}
	return
}

// FindStringInSlice returns the index of the first matching string in the slice,
// or -1 if there are no matches.
func FindStringInSlice(slice []string, find string) int {
	for idx, str := range slice {
		if str == find {
			return idx
		}
	}
	return -1
}

func RemoveStringFromSlice(slice []string, s string) (result []string) {
	for _, item := range slice {
		if item == s {
			continue
		}
		result = append(result, item)
	}
	return
}

// StringSetContains returns true if a set includes an item.
func StringSetContains(s []string, str string) bool {
	return (FindStringInSlice(s, str) >= 0)
}

func IsEqualStrings(a1 []string, a2 []string) bool {
	sort.Strings(a1)
	sort.Strings(a2)
	if len(a1) == len(a2) {
		for i, v := range a1 {
			if v != a2[i] {
				return false
			}
		}
	} else {
		return false
	}

	return true
}

func ContainsEqualFoldSubstring(str, substr string) bool {
	x := strings.ToLower(str)
	y := strings.ToLower(substr)
	return strings.Contains(x, y)
}

func GetNetAppVolumeCondition(
	volume *quarkv1alpha1.NetAppVolume,
	conditionType quarkv1alpha1.NetAppVolumeConditionType,
) *quarkv1alpha1.NetAppVolumeCondition {
	var condition *quarkv1alpha1.NetAppVolumeCondition
	for _, cond := range volume.Status.Conditions {
		if cond.Type == conditionType {
			condition = &cond
			break
		}
	}
	return condition
}

// Create a selector that finds data PVCs for a netappObj
func GetDataPvcSelector(storagePool *quarkv1alpha1.NetAppStoragePool) labels.Selector {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.TypeLabel] = "data"
	if storagePool != nil {
		selectorSet[quarkconfig.StoragePoolLabel] = storagePool.Name
	}
	return labels.SelectorFromSet(selectorSet)
}

func IsSMBConfigurationEqual(volume *quarkv1alpha1.NetAppVolume) bool {
	return (IsEqualSMBActiveDirectoryConfiguration(volume) &&
		IsEqualSMBShareProperties(volume))
}

func IsSMBServerConfigurationEqual(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	errConditionPresent, _ := IsErrorConditionPresent(storagePool.Status.Conditions,
		quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
	return (storagePool.Spec.SMBServerConfiguration == storagePool.Status.SMBServerConfiguration &&
		!errConditionPresent)
}

// ValidateIPv4 returns nil if the given IP address is a valid IPv4 address otherwise it returns error string comprising of invalid IP addresses
func ValidateIPv4(IPs []string) error {
	invalidIPs := ""
	failedReason := "Invalid IPv4 Address"
	for _, ip := range IPs {
		if net.ParseIP(ip) == nil || net.ParseIP(ip).To4() == nil {
			if invalidIPs == "" {
				invalidIPs = ip
			} else {
				invalidIPs = fmt.Sprintf("%s, %s", invalidIPs, ip)
				failedReason = "Invalid IPv4 Addresses"
			}
		}
	}
	if invalidIPs != "" {
		return fmt.Errorf("%s: %s", failedReason, invalidIPs)
	}
	return nil
}

// IsStoragePoolSMBConfigured returns "true" if SMB server id configured in storagepool.
// Tells whether user has started configuring SMB
func IsStoragePoolSMBConfigured(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	if storagePool.Spec.SMBServerConfiguration.Netbios != "" && storagePool.Spec.SMBServerConfiguration.Domain != "" {
		return true
	}
	return false
}

// IsStoragePoolSMBInitialized returns true if SMB is initialized for a storage pool
func IsStoragePoolSMBInitialized(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	if storagePool.Status.SMBServerConfiguration.Netbios != "" && storagePool.Status.SMBServerConfiguration.Domain != "" {
		return true
	}
	return false
}

// IsStoragePoolKerberosConfigured returns "true" if kerberos is configured in storagepool.
// Tells whether user has started configuring kerberos
func IsStoragePoolKerberosConfigured(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return storagePool.Spec.KerbServerConfiguration != nil &&
		storagePool.Spec.KerbServerConfiguration.ADServerName != "" &&
		storagePool.Spec.KerbServerConfiguration.KDCServerIP != ""
}

// IsStoragePoolLdapReady returns true if LDAP is ready for a storage pool
func IsStoragePoolLdapReady(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return (storagePool.Spec.LdapEnabled && storagePool.Spec.LdapEnabled == storagePool.Status.LdapEnabled &&
		storagePool.Spec.LDAPUpdateCount == storagePool.Status.LDAPUpdateCount)
}

func GetVolCreateADSecretName(volumeName string) string {
	return volumeName + "ad-create"
}

func GetVolModifyADSecretName(volumeName string) string {
	return volumeName + "ad-modify"
}

func GetVolDeleteADSecretName(volumeName string) string {
	return volumeName + "ad-delete"
}

func GetSMBServerModifySecretName(name string) string {
	return name + "smb-modify"
}

func GetVolPwdResetADSecretName(volumeName string) string {
	return volumeName + "ad-reset"
}

func CcpdFeatureEnabled(quark *quarkv1alpha1.Quark, storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return (storagePool.Spec.LdapEnabled || quark.Spec.EnableNfs41 || quark.Spec.EnableBackupSecurityUsers)
}

// GetContainerSpec gets the container spec by containerName.
func GetContainerSpec(containers []corev1.Container, containerName string) *corev1.Container {
	for i, c := range containers {
		if containerName == c.Name {
			return &containers[i]
		}
	}
	return nil
}

func MakeResolvConfString(dnsIPs []string) string {
	resolvData := ""
	for _, dnsIP := range dnsIPs {
		resolvData += "nameserver " + dnsIP + "\n"
	}
	return resolvData
}

func RoundUp(value float64, multiple int64) int64 {
	fm := float64(multiple)
	return int64(math.Ceil(value/fm) * fm)
}

// Allowed characters in kubernetes object names are alphanumeric, "-", and ".".
// K8s zones are of the form "<geo>-<subregion>-<zoneNumber>". Use "." as a
// separator, since the string is used to name storage classes and we need
// to be able to easily undo the join.
func GetFailureDomainString(zones []string) string {
	sort.Strings(zones)
	return strings.Join(zones, ".")
}

func GetFailureDomainZones(fd string) []string {
	return strings.Split(fd, ".")
}

func GetSCNameForFailureDomain(baseName, failureDomain string) string {
	if failureDomain == "" {
		return baseName
	}
	return fmt.Sprintf("%s-%s", baseName, failureDomain)
}

// If a node becomes unreachable, the pod's Ready status will be set to ConditionUnknown,
// but the pod itself will still be in the Running phase. In general, it seems like
// once a pod enters the Running phase, the phase doesn't change. To check if the pod is
// truly running, we check against the Phase and PodReady condition.
func IsPodRunning(pod *corev1.Pod) bool {
	var podReady = false
	for _, cond := range pod.Status.Conditions {
		if cond.Type == corev1.PodReady {
			podReady = cond.Status == corev1.ConditionTrue
			break
		}
	}
	return pod.Status.Phase == corev1.PodRunning && podReady
}

// We should consider a storage pool running iff dmap and Podrick are running,
// since those are necessary to serve data and ensure the SP responds to control
// plane functionality. All other components are ancillary, and the storage pool
// can function without them.
func SPPodRunning(spPod *corev1.Pod) bool {
	podrickRunning := false
	dmapRunning := false
	for _, containerStatus := range spPod.Status.ContainerStatuses {
		switch containerStatus.Name {
		case quarkconfig.QuarkContainerDMAP:
			dmapRunning = containerStatus.Ready
		case quarkconfig.QuarkContainerPodrick:
			podrickRunning = containerStatus.Ready
		}
	}
	return podrickRunning && dmapRunning
}

func PrepullContainersFromImageMap(data map[string]string, cloud string) []corev1.Container {
	prepullContainers := []corev1.Container{}
	for k, v := range data {
		container := corev1.Container{
			Name:  k,
			Image: v,
			// available only when shareProcessNamespace field is set
			Command: []string{"/proc/1/root/pause"},
		}

		container.SecurityContext = quarkconfig.GetContainerSecurityContext(quarkconfig.DefaultPrepullDaemonSet)
		container.SecurityContext.Capabilities.Add = []corev1.Capability{"SYS_PTRACE"}

		prepullContainers = append(prepullContainers, container)
	}
	return prepullContainers
}

func IsEqualSMBShareProperties(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Spec.SMBConfiguration.EncryptionEnabled == volume.Status.SMBConfiguration.EncryptionEnabled &&
		volume.Spec.SMBConfiguration.HideShare == volume.Status.SMBConfiguration.HideShare &&
		volume.Spec.SMBConfiguration.ShareAccessBasedEnumeration == volume.Status.SMBConfiguration.ShareAccessBasedEnumeration)

}

func IsEqualSMBActiveDirectoryConfiguration(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Spec.SMBConfiguration.Netbios == volume.Status.SMBConfiguration.Netbios &&
		volume.Spec.SMBConfiguration.Domain == volume.Status.SMBConfiguration.Domain &&
		volume.Spec.SMBConfiguration.EncryptedADCommunication == volume.Status.SMBConfiguration.EncryptedADCommunication &&
		volume.Spec.SMBConfiguration.PreferredADSite == volume.Status.SMBConfiguration.PreferredADSite &&
		IsEqualStrings(volume.Spec.SMBConfiguration.DNSIPs, volume.Status.SMBConfiguration.DNSIPs))

}

func NodeSelectorMapToString(nodeSelector map[string]string) (out string) {
	var keys []string
	for k := range nodeSelector {
		keys = append(keys, k)
	}
	sort.Strings(keys)
	for _, key := range keys {
		if out == "" {
			out = fmt.Sprintf("%s=%s", key, nodeSelector[key])
		} else {
			out = fmt.Sprintf("%s,%s=%s", out, key, nodeSelector[key])
		}
	}
	return
}

func NodeSelectorStringToMap(nodeSelector string) map[string]string {
	out := make(map[string]string)
	if len(nodeSelector) == 0 {
		return out
	}
	kvPairs := strings.Split(nodeSelector, ",")
	for _, pair := range kvPairs {
		kAndV := strings.Split(pair, "=")
		if len(kAndV) != 2 {
			log.WithFields(log.Fields{
				"nodeSelector": nodeSelector,
				"kAndV":        kAndV,
			}).Error("Invalid k/v pair")
			continue
		}
		out[kAndV[0]] = kAndV[1]
	}
	return out
}

func GetZonesForNodes(nodes []*corev1.Node) []string {
	zoneMap := make(map[string]bool)
	for _, node := range nodes {
		if zone, ok := node.Labels[quarkconfig.ZoneTopologyKey]; ok {
			zoneMap[zone] = true
		}
	}
	zones := make([]string, 0, len(zoneMap))
	for zone := range zoneMap {
		zones = append(zones, zone)
	}
	// This is needed to ensure stability when determining things like zone pairs.
	sort.Strings(zones)
	return zones
}

func GetResources(quarkPodBucket quarkv1alpha1.QuarkPodBucket) corev1.ResourceList {
	cpuRequest := resource.NewQuantity(0, resource.DecimalSI)
	memRequest := resource.NewQuantity(0, resource.BinarySI)
	for _, resources := range quarkPodBucket.ContainerResources {
		cpuRequest.Add(resources.CPUsRequest)
		memRequest.Add(resources.MemRequest)
	}
	return corev1.ResourceList{
		corev1.ResourceCPU:    *cpuRequest,
		corev1.ResourceMemory: *memRequest,
	}
}

// Returns the nearest power of 2 to the number rounded up or down
func NearestPowerofTwo(num float64, roundingfunction string) float64 {
	if roundingfunction == quarkconfig.CeilRoundFunction {
		return math.Pow(2, math.Ceil(math.Log2(num)))
	}
	return math.Pow(2, math.Floor(math.Log2(num)))
}

// Given a range of numbers , return all the numbers in the range as a string slice
func MakeRange(min, max int) []string {
	numList := make([]string, max-min+1)
	for i := range numList {
		numList[i] = fmt.Sprint((min + i))
	}
	return numList
}

func IsVersioned1_0Volume(volume *quarkv1alpha1.NetAppVolume) bool {
	if volume.Status.QubeVersion == "1.0" && volume.Status.Size.Value() != 0 {
		return true
	}
	return false
}

// Obtain QoS limits based on service-level requested
func GetServiceLevelLimits(
	svl string,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (IOPSPerTiB, throughput int, diskIOPSPerGiB, iopsLimit, tputLimit int, ampFactor float32, err error) {
	switch quarkconfig.ServiceLevel(svl) {
	case quarkconfig.ServiceLevelStandard, quarkconfig.ServiceLevelRegionalStandard:
		IOPSPerTiB = serviceLevel.Spec.IOPSPerTiB
		diskIOPSPerGiB = serviceLevel.Spec.DiskIOPSPerGiB
		throughput = serviceLevel.Spec.MiBsPerTiB
		ampFactor = serviceLevel.Spec.TpAmpFactor
		iopsLimit = serviceLevel.Spec.IOPSLimit
		tputLimit = serviceLevel.Spec.TputLimit
		return
	}
	err = fmt.Errorf("invalid ServiceLevel %v", serviceLevel)
	return
}

func GetContainerResourcesMap(qpb *quarkv1alpha1.QuarkPodBucket) map[string]quarkv1alpha1.ContainerResource {
	containerResourceMap := make(map[string]quarkv1alpha1.ContainerResource)
	if qpb == nil {
		return containerResourceMap
	}

	for _, containerResource := range qpb.ContainerResources {
		containerResourceMap[containerResource.ContainerName] = containerResource
	}

	return containerResourceMap
}

func baseInstancenameFromNode(name string) string {
	return name[0:strings.LastIndex(name, "-")]
}

func instanceGroupNameFromBaseInstance(name string) string {
	return fmt.Sprintf("%s-grp", name)
}

func GetMIGForNode(name string) string {
	return instanceGroupNameFromBaseInstance(baseInstancenameFromNode(name))
}

func GetVMRightSizingSelector(nodeSelectors map[string]string) (key, value string, ok bool) {
	vmrsKeys := []string{
		quarkconfig.OldInstanceTypeKey,
		quarkconfig.InstanceTypeKey,
	}
	for _, key := range vmrsKeys {
		value, ok := nodeSelectors[key]
		if ok {
			return key, value, true
		}
	}
	return "", "", false
}

// Given a full image path and a tag, update given image name and tag in the image path
// If Registry is given update registry
func UpdateTagInImage(imagePath string, imagenametag string, registry string) (string, error) {
	if registry != "" {
		// If registry is given we change the entire image string
		return registry + "/" + imagenametag, nil
	}

	// If registry is not given, we only change the image name and path
	imageSlice := strings.Split(imagePath, "/")
	if len(imageSlice) < 2 {
		return "", fmt.Errorf("Invalid image path %s", imagePath)
	}

	// Image Name:Tag will always be the last element
	imageSlice[len(imageSlice)-1] = imagenametag

	return strings.Join(imageSlice, "/"), nil
}

// Given an image path eg: us.gcr.io/netapp-hcl/podrick:kvalliyu-064d986e return us.gcr.io/netapp-hcl/
func GetRegistryFromImagePath(imagePath string) (string, error) {
	imagePathSlice := strings.Split(imagePath, "/")
	if len(imagePathSlice) < 2 {
		return "", fmt.Errorf("Invalid image path %s", imagePath)
	}

	// Leave out the last element in the slice, rest will be the registry path
	imagePathSlice = imagePathSlice[:len(imagePathSlice)-1]
	return strings.Join(imagePathSlice, "/"), nil
}

// getImageToken splits the string of a k8s "image" field by ":" or "@" (if @ is defined, meaning the image is in a digest string format)
// it returns the index passed as idx of the split, where int(0) would return the registry and int(1) would return the image tag/digest
// if the image string is in an invalid format it will return an error
func getImageToken(imagePath string, idx int) (string, error) {
	var imagePathSlice []string
	if strings.Contains(imagePath, "@") {
		imagePathSlice = strings.Split(imagePath, "@")
	} else {
		imagePathSlice = strings.Split(imagePath, ":")
	}

	if len(imagePathSlice) < 2 {
		return "", fmt.Errorf("invalid image path %s", imagePath)
	}

	return imagePathSlice[idx], nil
}

// GetImageTagOrDigest returns the image tag or digest from a image string
// For example: us.gcr.io/netapp-hcl/podrick:kvalliyu-064d986e would return kvalliyu-064d986e
// or us.gcr.io/netapp-hcl/podrick@sha256:1ab2c3fg would return sha256:1ab2c3fg
func GetImageTagOrDigest(imagePath string) (string, error) {
	return getImageToken(imagePath, 1)
}

// GetImageURI returns the an image string with the image tag/digest removed
// For example: us.gcr.io/netapp-hcl/podrick:kvalliyu-064d986e would return us.gcr.io/netapp-hcl/podrick
func GetImageURI(imagePath string) (string, error) {
	return getImageToken(imagePath, 0)
}

// Given a key and value, set the value to a field that matches the key in the struct
func SetStructField(obj interface{}, name string, value interface{}) error {
	structValue := reflect.ValueOf(obj).Elem()
	structFieldValue := structValue.FieldByName(name)

	if !structFieldValue.IsValid() {
		return fmt.Errorf("No such field: %s in obj", name)
	}

	if !structFieldValue.CanSet() {
		return fmt.Errorf("Cannot set %s field value", name)
	}

	structFieldType := structFieldValue.Type()
	val := reflect.ValueOf(value)
	if structFieldType != val.Type() {
		invalidTypeError := errors.New("Provided value type didn't match obj field type")
		return invalidTypeError
	}

	structFieldValue.Set(val)
	return nil
}

func FillStructFieldsFromMap(obj interface{}, mapVal map[string]interface{}) error {
	for k, v := range mapVal {
		err := SetStructField(obj, k, v)
		if err != nil {
			return err
		}
	}
	return nil
}

func IsNodeSchedulable(node corev1.Node) bool {
	if len(node.Spec.Taints) > 0 {
		for _, t := range node.Spec.Taints {
			if t.Effect == corev1.TaintEffectNoSchedule {
				return false
			}
		}
	}
	return true
}

func IsValidReplicationPolicy(replPolicyType quarkv1alpha1.NetAppVolumeReplicationPolicyType) error {
	if replPolicyType == quarkv1alpha1.ReplicationPolicyMirrorAllSnapshots ||
		replPolicyType == quarkv1alpha1.ReplicationPolicyMirrorAndVault ||
		replPolicyType == quarkv1alpha1.ReplicationPolicyMirrorLatest {
		return nil
	}
	return fmt.Errorf("invalid replication policy type %v. Valid policies include: %v, %v, %v",
		replPolicyType,
		quarkv1alpha1.ReplicationPolicyMirrorAllSnapshots,
		quarkv1alpha1.ReplicationPolicyMirrorAndVault,
		quarkv1alpha1.ReplicationPolicyMirrorLatest)
}

func IsValidReplicationSchedule(replPolicySchedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error {
	if replPolicySchedule == quarkv1alpha1.ReplicationSchedule10Minutely ||
		replPolicySchedule == quarkv1alpha1.ReplicationScheduleDaily ||
		replPolicySchedule == quarkv1alpha1.ReplicationScheduleHourly {
		return nil
	}
	return fmt.Errorf("invalid replication schedule %v. Valid schedule include: %v, %v, %v",
		replPolicySchedule,
		quarkv1alpha1.ReplicationSchedule10Minutely,
		quarkv1alpha1.ReplicationScheduleDaily,
		quarkv1alpha1.ReplicationScheduleHourly)
}

// CheckNfsVolume This function checks if the given volume is nfs3 or nfs4.1.
func CheckNfsVolume(volumeProtocol string) bool {
	protocols := strings.Split(volumeProtocol, ",")
	for protocolValues := range protocols {
		if protocols[protocolValues] == "cifs" || protocols[protocolValues] == "smb" {
			return false
		}
	}
	return true
}

func GetNHCHeadlessServiceFQDN(podName string, nhcNS string) string {
	return podName + "." + quarkconfig.NHCHeadlessService + "." + nhcNS + quarkconfig.NHCServiceDomain
}

func IsCMEKConfigured(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	_, labelPresent := storagePool.Labels[quarkconfig.CMEKConfigMapLabel]
	return labelPresent
}

func RemoveDuplicates(protocolArr []string) []string {
	protcolaMap := make(map[string]string)
	for _, protocols := range protocolArr {
		protcolaMap[protocols] = protocols
	}
	var uniqueProtocolArr []string
	for protocols, _ := range protcolaMap {
		uniqueProtocolArr = append(uniqueProtocolArr, protocols)
	}
	return uniqueProtocolArr
}

func ContainsNFSv4(protocols []string) bool {
	nfsVariations := []string{"nfs4.1", "nfsv4.1", "nfs4", "nfsv4"}
	for _, p := range protocols {
		p = strings.ToLower(p)
		for _, v := range nfsVariations {
			if p == v {
				return true
			}
		}
	}
	return false
}

func ContainsNFSv3(protocols []string) bool {
	for _, p := range protocols {
		p = strings.ToLower(p)
		if p == "nfs3" {
			return true
		}
	}
	return false
}

func ContainsBothNFS(protocols []string) bool {
	return ContainsNFSv4(protocols) && ContainsNFSv3(protocols)
}

// GetContainerTrustedCACertFilePath return quark trusted ca cert file path within the container
func GetContainerTrustedCACertFilePath() string {
	return filepath.Join(quarkconfig.QuarkTrustedCACertsPath, quarkconfig.CACertificateFileName)
}

// GetPodSpecVolumeForQuarkTrustedCACert returns secrets based volume obj for trusted ca certificate from secrets
func GetPodSpecVolumeForQuarkTrustedCACert() corev1.Volume {
	return corev1.Volume{
		Name: quarkconfig.QuarkTrustedCAPodVolumeName,
		VolumeSource: corev1.VolumeSource{
			Secret: &corev1.SecretVolumeSource{
				SecretName: quarkconfig.SecretNameQuarkTrustedCA,
			},
		},
	}
}

// GetContainerVolumeMountForQuarkTrustedCACert returns container volumeMount object for trusted ca certificat
func GetContainerVolumeMountForQuarkTrustedCACert() corev1.VolumeMount {
	return corev1.VolumeMount{
		Name:      quarkconfig.QuarkTrustedCAPodVolumeName,
		ReadOnly:  true,
		MountPath: quarkconfig.QuarkTrustedCACertsPath,
	}
}

// GetContainerVolumeMountForOSDefaultTrustedCACert returns container volumeMount object for trusted ca certificat
func GetContainerVolumeMountForOSDefaultTrustedCACert() corev1.VolumeMount {
	return corev1.VolumeMount{
		Name:      quarkconfig.QuarkTrustedCAPodVolumeName,
		ReadOnly:  true,
		MountPath: quarkconfig.OSDefaultTrustedCACertsPath,
	}
}

// AttachTrustedCACertVolumeToContainers creates volume for quark-trusted-ca secrets in the pod-spec and
// also create volume mount for specified containers
func AttachTrustedCACertVolumeToContainers(spec *corev1.PodSpec, containerNames []string) {
	if len(containerNames) == 0 {
		return
	}

	spec.Volumes = append(spec.Volumes, GetPodSpecVolumeForQuarkTrustedCACert())

	for _, containerName := range containerNames {
		containerSpec := GetContainerSpec(spec.Containers, containerName)
		containerSpec.VolumeMounts = append(containerSpec.VolumeMounts, GetContainerVolumeMountForQuarkTrustedCACert())
	}
}

func ValidateKerberosConfiguration(KerbServerConfiguration quarkv1alpha1.KerbServerConfiguration) error {
	if KerbServerConfiguration.KDCServerIP == "" || KerbServerConfiguration.ADServerName == "" {
		return fmt.Errorf("Kdcip and adservername can't be empty.")
	}

	err := ValidateIPv4([]string{KerbServerConfiguration.KDCServerIP})
	return err
}

func KerberosConfigurationCompareEqual(sp *quarkv1alpha1.NetAppStoragePool, isAbortOp bool) (bool, error) {
	if sp.Spec.KerbServerConfiguration == nil || sp.Status.KerbServerConfigurationStatus == nil {
		return false, fmt.Errorf("KerbServerConfiguration pointer is nil.")
	}

	if sp.Status.KerbServerConfigurationStatus.ADServerName != sp.Spec.KerbServerConfiguration.ADServerName ||
		sp.Status.KerbServerConfigurationStatus.KDCServerIP != sp.Spec.KerbServerConfiguration.KDCServerIP ||
		sp.Status.KerbConfigUpdateCount != sp.Spec.KerbConfigUpdateCount {
		return false, fmt.Errorf("Kerberos configuration is not reconciled yet.")
	}

	// Not checking SMB fields for Kerberos Abort as we support partial abort too.
	// SMB configurations is not aborted in partial abort case.
	if isAbortOp == false {
		// Netbios & domain are mandatory fields for SMB.
		if sp.Status.KerbServerConfigurationStatus.Netbios != sp.Spec.SMBServerConfiguration.Netbios ||
			sp.Status.KerbServerConfigurationStatus.Domain != sp.Spec.SMBServerConfiguration.Domain {
			return false, fmt.Errorf("Netbios/Domain for Kerberos is not reconciled yet.")
		}

		// Checking empty value as OU field is optional.
		if sp.Spec.SMBServerConfiguration.OrganizationUnit != "" &&
			sp.Status.KerbServerConfigurationStatus.OrganizationUnit != sp.Spec.SMBServerConfiguration.OrganizationUnit {
			return false, fmt.Errorf("OU for Kerberos is not reconciled yet.")
		}
	}

	return true, nil
}

func KerberosConfigurationInitialized(sp *quarkv1alpha1.NetAppStoragePool) bool {
	if sp.Status.KerbServerConfigurationStatus != nil &&
		sp.Status.KerbServerConfigurationStatus.ADServerName != "" &&
		sp.Status.KerbServerConfigurationStatus.KDCServerIP != "" {
		return true
	}
	return false
}

func KerberosConfigurationErrorPresent(sp *quarkv1alpha1.NetAppStoragePool) (bool, string) {
	if errPresent, _ := IsErrorConditionPresent(sp.Status.Conditions,
		quarkv1alpha1.NetAppStoragePoolKerbConfigurationError); errPresent {
		err := GetNetAppStoragePoolCondition(sp, quarkv1alpha1.NetAppStoragePoolKerbConfigurationError).Reason
		return true, err
	}
	if errPresent, _ := IsErrorConditionPresent(sp.Status.Conditions,
		quarkv1alpha1.NetAppStoragePoolKerbConfigurationRetryError); errPresent {
		err := GetNetAppStoragePoolCondition(sp, quarkv1alpha1.NetAppStoragePoolKerbConfigurationRetryError).Reason
		return true, err
	}
	return false, ""
}

// returns true/false, if an item is found/not-found in the group
func Contains(item string, group []string) bool {
	for _, item_found := range group {
		if item_found == item {
			return true
		}
	}
	return false
}
